classdef (Abstract) Body < handle
	%Body A rigid body connected through a joint to a parent
	
	%%
	properties
		name     % Optional name
		color    % Color for rendering
		axisSize % Axis size for rendering
		density  % Mass/volume
		I_i      % Inertia at body center
		E_ji     % Where the body is wrt joint
		E_ij     % Where the joint is wrt body
		E_wi     % Where the body is wrt world
		E_iw     % Where the world is wrt body
		E_ip     % Where the parent is wrt body
		E_im     % Where the mesh is wrt body (used only for display)
		Ad_ji    % Adjoint of E_ji
		Ad_ij    % Adjoint of E_ij
		Ad_iw    % Adjoint of E_iw
		Ad_wi    % Adjoint of E_wi
		Ad_ip    % Adjoint of E_ip
		Addot_wi % Adjoint dot of E_wi, used by the Jacobian
		phi      % Twist at body center
		damping  % Viscous damping
		fext     % External force in body space (3D position and 3D force)
		presc    % Prescribed motion constraint
		joint    % Joint to parent
		idxM     % Maximal indices
		next     % Next body in traversal order
	end
	
	%%
	methods
		%%
		function this = Body(density)
			global countB countCM CM;
			this.name = ['body',num2str(countB)];
			this.color = CM(mod(countCM-1,size(CM,1))+1,:);
			this.axisSize = 1.0;
			this.density = density;
			this.I_i = []; % will be set manually or computed by the subclass
			this.E_ji = eye(4);
			this.E_ij = eye(4);
			this.E_wi = eye(4);
			this.E_iw = eye(4);
			this.E_ip = eye(4);
			this.E_im = eye(4);
			this.Ad_ji = eye(6);
			this.Ad_ij = eye(6);
			this.Ad_iw = eye(6);
			this.Ad_wi = eye(6);
			this.Ad_ip = eye(6);
			this.Addot_wi = eye(6);
			this.phi = zeros(6,1);
			this.damping = 0;
			this.joint = {};
			countB = countB + 1;
			countCM = countCM + 1;
		end
		
		%%
		function setBodyTransform(this,E)
			% Sets the transform of this body wrt parent joint
			this.E_ji = E;
			this.E_ij = se3.inv(this.E_ji);
			this.Ad_ji = se3.Ad(this.E_ji);
			this.Ad_ij = se3.Ad(this.E_ij);
		end
		
		%%
		function setMeshTransform(this,E)
			% Sets the transform of the mesh wrt the body.
			% This does not change the dynamics. The mesh is only used for
			% display, and the inertia will need to be set manually.
			this.E_im = E;
		end
		
		%%
		function setDamping(this,damping)
			% Sets viscous damping
			this.damping = damping;
		end
		
		%%
		function countDofs(this)
			% Counts maximal DOFs
			nm = redmax.Scene.countM();
			this.idxM = nm + (1:6);
			nm = nm + 6;
			redmax.Scene.countM(nm);
		end
		
		%%
		function computeInertia(this)
			% Computes inertia at joint
			this.joint.computeInertia();
		end
		
		%%
		function update(this)
			% Updates this body's transforms and velocities
			this.E_wi = this.joint.E_wj*this.E_ji;
			this.E_iw = se3.inv(this.E_wi);
			this.Ad_wi = se3.Ad(this.E_wi);
			this.Ad_iw = se3.Ad(this.E_iw);
			this.E_ip = eye(4);
			if ~isempty(this.joint.parent)
				this.E_ip = this.E_iw*this.joint.parent.body.E_wi;
			end
			this.Ad_ip = se3.Ad(this.E_ip);
			% Body velocity
			this.phi = this.Ad_ij*this.joint.V;
			this.Addot_wi = se3.Addot(this.E_wi,this.phi);
		end
		
		%%
		function updateQuick(this)
			% Updates this body's transforms
			this.E_wi = this.joint.E_wj*this.E_ji;
		end
		
		%%
		function [Mm,fm,Km,Dm] = computeMassGrav(this,grav,Mm,fm,Km,Dm)
			% Computes maximal mass matrix and force vector
			nm = redmax.Scene.countM();
			if nargout == 2
				if nargin == 2
					Mm = zeros(nm,nm);
					fm = zeros(nm,1);
				end
			else
				if nargin == 2
					Mm = zeros(nm,nm);
					fm = zeros(nm,1);
					Km = zeros(nm,nm);
					Dm = zeros(nm,nm);
				end
			end
			% Coriolis and gravity
			rows = this.idxM;
			M_i = diag(this.I_i);
			Mm(rows,rows) = M_i;
			adt = se3.ad(this.phi)';
			fcor = adt*M_i*this.phi;
			R_wi = this.E_wi(1:3,1:3);
			R_iw = R_wi';
			fgrav = zeros(6,1);
			mass = M_i(4,4);
			grav_i = R_iw*grav;
			fgrav(4:6) = mass*grav_i; % wrench in body space
			fm(rows) = fm(rows) + fcor + fgrav;
			if nargout == 2
				% Go to the next body
				if ~isempty(this.next)
					[Mm,fm] = this.next.computeMassGrav(grav,Mm,fm);
				end
			else
				% Derivative of gravity wrt position
				Km(rows(4:6),rows(1:3)) = Km(rows(4:6),rows(1:3)) + se3.brac(fgrav(4:6));
				% Derivative of Coriolis wrt velocity
				e1 = se3.brac([1 0 0]);
				e2 = se3.brac([0 1 0]);
				e3 = se3.brac([0 0 1]);
				z3 = [0 0 0]';
				Iw = this.I_i(1:3).*this.phi(1:3);
				mv = mass*this.phi(4:6);
				Dm(rows,rows) =  Dm(rows,rows) + adt*M_i - [
					e1*Iw, e2*Iw, e3*Iw, e1*mv, e2*mv, e3*mv
					e1*mv, e2*mv, e3*mv,    z3,    z3,    z3];
				% Go to the next body
				if ~isempty(this.next)
					[Mm,fm,Km,Dm] = this.next.computeMassGrav(grav,Mm,fm,Km,Dm);
				end
			end
		end
		
		%%
		function [f,D] = computeForceDamping(this,f,D)
			% Computes maximal damping force vector and matrix
			nm = redmax.Scene.countM();
			if nargin == 1
				f = zeros(nm,1);
				D = zeros(nm);
			elseif nargin == 2
				D = zeros(nm);
			end
			rows = this.idxM;
			if this.damping > 0
				fi = -this.damping*this.phi;
				Di = this.damping*eye(6);
				f(rows) = f(rows) + fi;
				D(rows,rows) = D(rows,rows) + Di;
			end
			if ~isempty(this.fext)
				G = se3.Gamma(this.fext.x);
				f(rows) = f(rows) + G'*this.fext.v;
			end
			if ~isempty(this.next)
				[f,D] = this.next.computeForceDamping(f,D);
			end
		end
		
		%%
		function [T,V] = computeEnergies(this,grav,T,V)
			% Computes kinetic and potential energies
			% The kinetic energy can also be computed by the joint as
			% T = 0.5*this.V'*this.I_j*this.V;
			T = T + 0.5*this.phi'*diag(this.I_i)*this.phi;
			V = V - this.I_i(6)*grav'*this.E_wi(1:3,4);
		end
		
		%%
		function draw(this)
			s = this.getAxisSize();
			if s > 0
				xlines = zeros(3,0);
				ylines = zeros(3,0);
				zlines = zeros(3,0);
				[x,y,z] = se3.drawAxis(this.E_wi,s); % body frame
				xlines = [xlines,x,nan(3,1)];
				ylines = [ylines,y,nan(3,1)];
				zlines = [zlines,z,nan(3,1)];
				[x,y,z] = se3.drawAxis(this.E_wi*this.E_im,s); % mesh frame
				xlines = [xlines,x,nan(3,1)];
				ylines = [ylines,y,nan(3,1)];
				zlines = [zlines,z,nan(3,1)];
				plot3(xlines(1,:),xlines(2,:),xlines(3,:),'r','LineWidth',1);
				plot3(ylines(1,:),ylines(2,:),ylines(3,:),'g','LineWidth',1);
				plot3(zlines(1,:),zlines(2,:),zlines(3,:),'b','LineWidth',1);
			end
			%text(this.E_wi(1,4),this.E_wi(2,4),this.E_wi(3,4),this.name,'Color','r');
			this.draw_();
			if ~isempty(this.fext)
				% Draw external force
				x = this.E_wi*[this.fext.x;1];
				v = this.E_wi*[this.fext.v;0];
				xv = [x,x+2e-4*v];
				%plot3(xv(1,:),xv(2,:),xv(3,:),'k-','LineWidth',2);
			end
			if ~isempty(this.next)
				this.next.draw();
			end
		end

		%%
		function setColor(this,c)
			this.color = c;
		end
		
		%%
		function setAxisSize(this,s)
			this.axisSize = s;
		end
		
		%%
		function s = getAxisSize(this)
			s = this.axisSize;
		end
	end
	
	%%
	methods (Abstract)
		%%
		draw_(this)
		computeInertia_(this)
	end
end
