classdef (Abstract) Joint < handle
	%Joint A generic joint between two bodies
	%   A joint is defined between a parent and the child. The DOF, q, of
	%   the joint is the relative displacement of the child wrt the parent.
	
	%%
	properties
		name     % Optional name
		body     % Attached body
		parent   % Parent joint
		children % Children joints
		ndof     % Number of DOF
		q        % Position
		qdot     % Velocity
		q0       % Last position
		qdot0    % Last velocity
		qInit    % Initial position (for joint stiffness)
		tau      % Joint torque
		Kr       % Joint stiffness
		Dr       % Joint damping
		S        % Jacobian
		Sdot     % dS/dt
		I_j      % Inertia at the joint
		V        % Twist at parent joint
		E_pj     % Transform of this joint wrt parent joint
		E_pj0    % Transform when q is zero (initial transform)
		E_jp     % Transform of parent joint wrt this joint
		Ad_jp    % Adjoint of E_jp
		E_wj     % Transform of this joint wrt world
		E_wj0    % Transform of this joint wrt world (initial transform)
		qLower   % Lower limit
		qUpper   % Upper limit
		next     % Forward recursive ordering
		prev     % Reverse recursive ordering
		idxR     % Reduced indices
		axisSize % Display axis size
	end
	
	%%
	properties (Access = protected)
		Q % Transformation matrix applied about the joint
	end
	
	%%
	methods
		%%
		function this = Joint(parent,body,ndof)
			if isempty(parent)
				this.name = ['NULL-',body.name];
			else
				this.name = [parent.body.name,'-',body.name];
			end
			this.parent = parent;
			this.body = body;
			this.ndof = ndof;
			this.q = zeros(ndof,1);
			this.qdot = zeros(ndof,1);
			this.q0 = zeros(ndof,1);
			this.qdot0 = zeros(ndof,1);
			this.tau = zeros(ndof,1);
			this.Kr = 0;
			this.Dr = 0;
			this.S = zeros(6,ndof);
			this.Sdot = zeros(6,ndof);
			this.I_j = eye(6);
			this.V = zeros(6,1);
			this.E_pj0 = eye(4);
			this.qLower = -inf(ndof,1);
			this.qUpper = inf(ndof,1);
			body.joint = this;
			if ~isempty(parent)
				parent.children{end+1} = this;
			end
			this.idxR = 0;
			this.axisSize = 1.0;
		end
		
		%%
		function setJointTransform(this,E)
			% Sets the transform of this joint wrt parent joint
			this.E_pj0 = E;
		end
		
		%%
		function setStiffness(this,stiffness)
			% Sets this joint's linear stiffness
			this.Kr = stiffness;
		end
		
		%%
		function setDamping(this,damping)
			% Sets this joint's linear velocity damping
			this.Dr = damping;
		end

		%%
		function setLimits(this,lower,upper)
			this.qLower = lower;
			this.qUpper = upper;
		end
		
		%%
		function order = getTraversalOrder(this,order)
			% Gets traversal order. Given the tree structure (parent,
			% children) of the joints, this function returns an array
			% representation so that the parent comes before all of its
			% children.
			if nargin == 1
				order = [];
			end
			order(end+1) = length(order) + 1;
			for i = 1 : length(this.children)
				order = this.children{i}.getTraversalOrder(order);
			end
		end
		
		%%
		function countDofs(this)
			% Counts reduced DOFs
			nr = redmax.Scene.countR();
			this.idxR = nr + (1:this.ndof);
			nr = nr + this.ndof;
			redmax.Scene.countR(nr);
			this.body.countDofs();
			% Use the current q as the initial q
			this.qInit = this.q;
		end

		%%
		function [qAll,qdotAll] = getQ(this,qAll,qdotAll)
			% Gathers q (and optionally qdot) into (a) global vector(s)
			nr = redmax.Scene.countR();
			if nargin == 1
				qAll = zeros(nr,1);
				if nargout == 2
					qdotAll = zeros(nr,1);
				end
			end
			qAll(this.idxR) = this.q;
			if nargout == 2
				qdotAll(this.idxR) = this.qdot;
			end
			% Go to the next joint
			if ~isempty(this.next)
				if nargout == 1
					qAll = this.next.getQ(qAll);
				else
					[qAll,qdotAll] = this.next.getQ(qAll,qdotAll);
				end
			end
		end
		
		%%
		function qdotAll = getQdot(this,qdotAll)
			% Gathers qdot into a global vector
			nr = redmax.Scene.countR();
			if nargin == 1
				qdotAll = zeros(nr,1);
			end
			qdotAll(this.idxR) = this.qdot;
			% Go to the next joint
			if ~isempty(this.next)
				qdotAll = this.next.getQdot(qdotAll);
			end
		end
		
		%%
		function setQ(this,qAll,qdotAll)
			% Scatters q (and optionally qdot) from (a) global vector(s)
			this.q(1:this.ndof) = qAll(this.idxR);
			if nargin == 3
				this.qdot(1:this.ndof) = qdotAll(this.idxR);
			end
			% Go to the next joint
			if ~isempty(this.next)
				if nargin == 2
					this.next.setQ(qAll);
				else
					this.next.setQ(qAll,qdotAll);
				end
			end
		end
		
		%%
		function setQdot(this,qdotAll)
			% Scatters qdot from a global vector
			this.qdot(1:this.ndof) = qdotAll(this.idxR);
			% Go to the next joint
			if ~isempty(this.next)
				this.next.setQdot(qdotAll);
			end
		end
		
		%%
		function [q0All,qdot0All] = getQ0(this,q0All,qdot0All)
			% Gathers q0 (and optionally qdot0) into (a) global vector(s)
			nr = redmax.Scene.countR();
			if nargin == 1
				q0All = zeros(nr,1);
				if nargout == 2
					qdot0All = zeros(nr,1);
				end
			end
			q0All(this.idxR) = this.q0;
			if nargout == 2
				qdot0All(this.idxR) = this.qdot0;
			end
			% Go to the next joint
			if ~isempty(this.next)
				if nargout == 1
					q0All = this.next.getQ0(q0All);
				else
					[q0All,qdot0All] = this.next.getQ0(q0All,qdot0All);
				end
			end
		end
		
		%%
		function qdot0All = getQdot0(this,qdot0All)
			% Gathers qdot0 into a global vector
			nr = redmax.Scene.countR();
			if nargin == 1
				qdot0All = zeros(nr,1);
			end
			qdot0All(this.idxR) = this.qdot0;
			% Go to the next joint
			if ~isempty(this.next)
				qdot0All = this.next.getQdot0(qdot0All);
			end
		end
		
		%%
		function setQ0(this,q0All,qdot0All)
			% Scatters q0 (and optionally qdot0) from (a) global vector(s)
			this.q0(1:this.ndof) = q0All(this.idxR);
			if nargin == 3
				this.qdot0(1:this.ndof) = qdot0All(this.idxR);
			end
			% Go to the next joint
			if ~isempty(this.next)
				if nargin == 2
					this.next.setQ0(q0All);
				else
					this.next.setQ0(q0All,qdot0All);
				end
			end
		end
		
		%%
		function setQdot0(this,qdot0All)
			% Scatters qdot0 from a global vector
			this.qdot0(1:this.ndof) = qdot0All(this.idxR);
			% Go to the next joint
			if ~isempty(this.next)
				this.next.setQdot0(qdot0All);
			end
		end
		
		%%
		function computeInertia(this)
			% Fix inertia matrix shape if necessary
			% TODO: do this in the Body class
			I_i = this.body.I_i;
			if numel(I_i) ~= 6 % matrix
				I_i = diag(I_i);
			end
			if size(I_i,1) == 1 % row vector
				I_i = I_i';
			end
			this.body.I_i = I_i;
			m = this.body.I_i(4);
			R = this.body.E_ji(1:3,1:3);
			p = this.body.E_ji(1:3,4);
			pBrac = se3.brac(p);
			Ic = diag(this.body.I_i(1:3));
			this.I_j = [R*Ic*R'+m*(pBrac'*pBrac),m*pBrac;m*pBrac',m*eye(3)];
			% which can also be computed as:
			%   this.I_j = this.body.Ad_ij'*diag(this.body.I_i)*this.body.Ad_ij;
		end
		
		%%
		function update(this)
			% Updates this joint and the attached body
			this.update_();
			% Transforms and adjoints
			if isempty(this.E_pj0)
				this.E_pj = this.Q;
			else
				this.E_pj = this.E_pj0*this.Q;
			end
			this.E_jp = se3.inv(this.E_pj);
			this.Ad_jp = se3.Ad(this.E_jp);
			if isempty(this.parent)
				E_wp = eye(4);
			else
				E_wp = this.parent.E_wj;
			end
			this.E_wj = E_wp*this.E_pj;
			this.E_wj0 = E_wp*this.E_pj0;
			% Joint velocity
			if this.ndof == 0
				this.V = zeros(6,1);
			else
				this.V = this.S*this.qdot;
			end
			if ~isempty(this.parent) % Add parent velocity
				this.V = this.V + this.Ad_jp*this.parent.V;
			end
			% Update attached body
			this.body.update();
			if ~isempty(this.next)
				this.next.update();
			end
		end
		
		%%
		function updateQuick(this)
			% Updates this joint and the attached body
			this.updateQuick_();
			% Transforms
			if isempty(this.E_pj0)
				this.E_pj = this.Q;
			else
				this.E_pj = this.E_pj0*this.Q;
			end
			if isempty(this.parent)
				E_wp = eye(4);
			else
				E_wp = this.parent.E_wj;
			end
			this.E_wj = E_wp*this.E_pj;
			this.E_wj0 = E_wp*this.E_pj0;
			% Update attached body
			this.body.updateQuick();
			if ~isempty(this.next)
				this.next.updateQuick();
			end
		end
		
		%%
		function [fr,Kr] = computeForceStiffness(this,fr,Kr)
			% Computes joint stiffness force vector and matrix
			nr = redmax.Scene.countR();
			if nargin == 1
				fr = zeros(nr,1);
				Kr = zeros(nr);
			elseif nargin == 2
				Kr = zeros(nr);
			end
			rows = this.idxR;
			% We'll add the joint torque here rather than having a
			% separate function.
			fr(rows) = fr(rows) + this.tau - this.Kr*(this.q - this.qInit);
			Kr(rows,rows) = Kr(rows,rows) - this.Kr*eye(this.ndof);
			if ~isempty(this.next)
				[fr,Kr] = this.next.computeForceStiffness(fr,Kr);
			end
		end
		
		%%
		function [fr,Dr] = computeForceDamping(this,fr,Dr)
			% Computes joint damping force vector and matrix
			nr = redmax.Scene.countR();
			if nargin == 1
				fr = zeros(nr,1);
				Dr = zeros(nr);
			elseif nargin == 2
				Dr = zeros(nr);
			end
			rows = this.idxR;
			fr(rows) = fr(rows) - this.Dr*this.qdot;
			Dr(rows,rows) = Dr(rows,rows) - this.Dr*eye(this.ndof);
			if ~isempty(this.next)
				[fr,Dr] = this.next.computeForceDamping(fr,Dr);
			end
		end
		
		%%
		function [J,Jdot] = computeJacobian(this,J,Jdot)
			% Computes the redmax Jacobian
			nm = redmax.Scene.countM();
			nr = redmax.Scene.countR();
			if nargout == 1
				% Only compute J
				if nargin == 1
					J = zeros(nm,nr);
				end
				rows = this.body.idxM;
				cols = this.idxR;
				Ad_ij = this.body.Ad_ij;
				J(rows,cols) = Ad_ij*this.S;
				% Loop through all ancestors
				jointA = this.parent;
				while ~isempty(jointA)
					rowsP = this.parent.body.idxM;
					Ad_ip = this.body.Ad_ip;
					colsA = jointA.idxR;
					J(rows,colsA) = Ad_ip*J(rowsP,colsA);
					jointA = jointA.parent;
				end
				if ~isempty(this.next)
					J = this.next.computeJacobian(J);
				end
			else
				% Compute J and Jdot
				if nargin == 1
					J = zeros(nm,nr);
					Jdot = zeros(nm,nr);
				end
				rows = this.body.idxM;
				cols = this.idxR;
				Ad_ij = this.body.Ad_ij;
				J(rows,cols) = Ad_ij*this.S;
				Jdot(rows,cols) = Ad_ij*this.Sdot;
				% Loop through all ancestors
				jointA = this.parent;
				while ~isempty(jointA)
					rowsP = this.parent.body.idxM;
					Ad_ip = this.body.Ad_ip;
					Ad_iw = this.body.Ad_iw;
					Ad_wp = this.parent.body.Ad_wi;
					Addot_wi = this.body.Addot_wi;
					Addot_wp = this.parent.body.Addot_wi;
					Addot_ip = -Ad_iw*(Addot_wi*Ad_iw*Ad_wp - Addot_wp);
					colsA = jointA.idxR;
					J(rows,colsA) = Ad_ip*J(rowsP,colsA);
					Jdot(rows,colsA) = Ad_ip*Jdot(rowsP,colsA) + Addot_ip*J(rowsP,colsA);
					jointA = jointA.parent;
				end
				if ~isempty(this.next)
					[J,Jdot] = this.next.computeJacobian(J,Jdot);
				end
			end
		end

		%%
		function [lb,ub] = computeConstraints(this,lb,ub)
			nr = redmax.Scene.countR();
			if nargin == 1
				lb = -inf(nr,1);
				ub = inf(nr,1);
			end
			rows = this.idxR;
			for k = 1 : this.ndof
				if this.q(k) < this.qLower(k)
					% Want lower bound to be positive for stabilization
					lb(rows(k)) = this.qLower(k) - this.q(k);
				end
				if this.q(k) > this.qUpper(k)
					% Want upper bound to be negative for stabilization
					ub(rows(k)) = this.qUpper(k) - this.q(k);
				end
			end
			if ~isempty(this.next)
				[lb,ub] = this.next.computeConstraints(lb,ub);
			end
		end
		
		%%
		function [T,V] = computeEnergies(this,grav,T,V)
			% Computes kinetic and potential energies
			if nargin == 2
				T = 0;
				V = 0;
			end
			[T,V] = this.body.computeEnergies(grav,T,V);
			dq = this.q - this.qInit;
			V = V + 0.5*this.Kr*(dq'*dq);
			if ~isempty(this.next)
				[T,V] = this.next.computeEnergies(grav,T,V);
			end
		end
		
		%%
		function draw(this)
			s = this.getAxisSize();
			if s > 0
				xlines = zeros(3,0);
				ylines = zeros(3,0);
				zlines = zeros(3,0);
				[x,y,z] = se3.drawAxis(this.E_wj0,s,2);
				xlines = [xlines,x,nan(3,1)];
				ylines = [ylines,y,nan(3,1)];
				zlines = [zlines,z,nan(3,1)];
				[x,y,z] = se3.drawAxis(this.E_wj,s,2);
				xlines = [xlines,x,nan(3,1)];
				ylines = [ylines,y,nan(3,1)];
				zlines = [zlines,z,nan(3,1)];
				plot3(xlines(1,:),xlines(2,:),xlines(3,:),'r','LineWidth',1);
				plot3(ylines(1,:),ylines(2,:),ylines(3,:),'g','LineWidth',1);
				plot3(zlines(1,:),zlines(2,:),zlines(3,:),'b','LineWidth',1);
			end
			%text(this.E_wj(1,4),this.E_wj(2,4),this.E_wj(3,4),this.name);
			this.draw_();
			if ~isempty(this.next)
				this.next.draw();
			end
		end
		
		%%
		function setAxisSize(this,s)
			this.axisSize = s;
		end
		
		%%
		function s = getAxisSize(this)
			s = this.axisSize;
		end
	end
	
	%%
	methods (Access = protected)
		
		%%
		function update_(this) %#ok<MANU>
		end
		
		%%
		function updateQuick_(this) %#ok<MANU>
		end
		
		%%
		function draw_(this) %#ok<MANU>
		end
	end
end
