classdef (Abstract) Muscle < handle
	%Muscle A muscle with mass points
	
	properties
		name       % Optional name
		as         % Mass point percentage location along length
		ms         % Mass at each mass point
		xs         % Mass point positions in world
		vs         % Mass point velocities in world
		mass       % Total mass
		bO         % Origin body
		xO         % Origin point
		bI         % Insertion body
		xI         % Insertion point
		xOw        % Current world position of the origin
		xIw        % Current world position of the origin
		useJdotFD  % Whether to use finite differencing for computing Jdot
		idxA       % Muscle point indices (row indices in Jacobian)
		next       % Next muscle in traversal order
		seed       % DEBUG: rng seed for emulating noisy NN
		mmm        % Millard muscle model
	end
	
	methods
		%%
		function this = Muscle(as,ms,bO,xO,bI,xI)
			global na
			if nargin == 0
				return
			end
			this.name = ['muscle',num2str(na)];
			n = min(length(as),length(ms));
			this.as = as(1:n);
			this.ms = ms(1:n);
			if n < 1
				fprintf('No mass points\n');
				n = 1;
			end
			this.xs = zeros(3,n);
			this.vs = zeros(3,n);
			this.mass = sum(this.ms);
			this.bO = bO;
			this.xO = xO;
			this.bI = bI;
			this.xI = xI;
			this.xOw = zeros(3,1);
			this.xIw = zeros(3,1);
			this.useJdotFD = false;
			this.seed = 0;
			this.mmm = [];
		end
		
		%%
		function countDofs(this)
			if isempty(this.as)
				return
			end
			n = size(this.xs,2);
			na = redmax.Scene.countA();
			this.idxA = na + (1:(3*n));
			na = na + 3*n;
			redmax.Scene.countA(na);
		end
		
		%%
		function update(this)
			% Update origin and insertion
			if ~isempty(this.xO)
				this.xOw = this.bO.E_wi(1:3,:)*[this.xO;1];
				this.xIw = this.bI.E_wi(1:3,:)*[this.xI;1];
			end
			% Update mass points
			this.update_();
			if ~isempty(this.next)
				this.next.update();
			end
		end

		%%
		function setDebugRngSeed(this,seed)
			this.seed = seed;
			if ~isempty(this.next)
				this.next.setDebugRngSeed(seed);
			end
		end
		
		%%
		function [Jam,dJam] = computeJacobian(this,Jam,dJam,e)
			if isempty(this.as)
				return
			end
			na = redmax.Scene.countA();
			nm = redmax.Scene.countM();
			if nargin <= 2
				Jam = zeros(na,nm);
				dJam = zeros(na,nm);
				e = 0;
			end
			[Jam,dJam] = this.computeJacobian_(Jam,dJam,e);
			if ~isempty(this.next)
				[Jam,dJam] = this.next.computeJacobian(Jam,dJam,e);
			end
		end
		
		%%
		function [JamNN,JarNN] = computeJacobianNN(this,JamNN,JarNN)
			if isempty(this.as)
				return
			end
			na = redmax.Scene.countA();
			nm = redmax.Scene.countM();
			nr = redmax.Scene.countR();
			if nargin <= 2
				JamNN = zeros(na,nm);
				JarNN = zeros(na,nr);
			end
			[JamNN,JarNN] = this.computeJacobianNN_(JamNN,JarNN);
			if ~isempty(this.next)
				[JamNN,JarNN] = this.next.computeJacobianNN(JamNN,JarNN);
			end
		end
		
		%%
		function Jar = computeJacobianFD(this,idxR,e,Jar)
			if isempty(this.as)
				return
			end
			na = redmax.Scene.countA();
			nr = redmax.Scene.countR();
			if nargin <= 1
				idxR = 0;
			end
			if nargin <= 2
				e = 0;
			end
			if nargin <= 3
				Jar = zeros(na,nr);
			end
			if idxR == 0
				% Fill the Jacobian with position data at current joint
				% angles.
				for i = 1 : nr
					Jar(this.idxA,i) = reshape(this.xs,[],1);
				end
			else
				% For column idxR, subtract the updated positions and
				% divide to compute finite difference.
				Jar(this.idxA,idxR) = (reshape(this.xs,[],1) - Jar(this.idxA,idxR))/e;
			end
			if ~isempty(this.next)
				Jar = this.next.computeJacobianFD(idxR,e,Jar);
			end
		end
		
		%%
		function [Ma,fa] = computeMassGrav(this,grav,Ma,fa)
			if nargin == 2
				na = redmax.Scene.countA();
				Ma = zeros(na,na);
				fa = zeros(na,1);
			end
			if isempty(this.as)
				return
			end
			n = size(this.xs,2);
			I = eye(3);
			for i = 1 : n
				idxAi = this.idxA(3*(i-1)+(1:3));
				mi = this.ms(i);
				Ma(idxAi,idxAi) = mi*I;
				fa(idxAi) = mi*grav;
			end
			if ~isempty(this.next)
				[Ma,fa] = this.next.computeMassGrav(grav,Ma,fa);
			end
		end

		%%
		function initHMM(this,lMopt,lTslack,fMopt)
			% Use default values for most parameters
			beta = 0.1; % damping
			%lMopt = 0.1; % optimal muscle length
			%lTslack = 0.2; % tendon slack length
			vMmax = 10; % maximum contraction velocity
			alphaMopt = 0.0; % pennation angle at optimal muscle length
			%fMopt = 100; % peak isometric force
			amin = 0.01; % minimum activation
			tauA = 0.01; % activation constant
			tauD = 0.4; % deactivation constant
			
			% Create muscle curves
			curveAFL = hmm.CurveActiveForceLength();
			curvePFL = hmm.CurveFiberForceLength();
			curveTFL = hmm.CurveTendonForceLength();
			curveFV = hmm.CurveForceVelocity();
			
			% Create muscle
			muscle = hmm.MuscleMillard2012Damped(beta,lMopt,lTslack,vMmax,alphaMopt,fMopt,amin,tauA,tauD,curveAFL,curvePFL,curveTFL,curveFV);
			this.mmm.muscle = muscle;
			
			% Initial activation
			muscle.act = 0.05; 
			this.mmm.exc = muscle.act;

			% Initial MT length
			% Compute initial muscle length based on initial a and lMT,
			% assuming vM is zero.
			lMT = lMopt + lTslack;
			this.mmm.lMT = lMT;
			muscle.initEquilibrium(lMT);
		end

		%%
		function setExcitationHMM(this,e)
			this.mmm.exc = e;
		end

		%%
		function fm = computeForceHMM(this,fm)
			if nargin == 1
				nm = redmax.Scene.countM();
				fm = zeros(nm,1);
			end
			if ~isempty(this.mmm)
				% Tendon force
				fT = this.mmm.muscle.fT;
				% Application points, directions, and associated bodies
				points = this.getForcePointsHMM();
				for i = 1 : length(points)
					xw = points(i).xw; % world
					dw = points(i).dw; % world
					fw = fT*dw;
					body = points(i).body;
					E = body.E_wi;
					R = E(1:3,1:3);
					xl = se3.inv(E)*[xw(1:3);1];
					G = se3.Gamma(xl);
					idx = body.idxM;
					fm(idx) = fm(idx) + G'*R'*fw;
				end
			end
			if ~isempty(this.next)
				fm = this.next.computeForceHMM(fm);
			end
		end

		%%
		function integrateHMM(this,h)
			if ~isempty(this.mmm)
				% TODO: implicit integration
				lMT = this.computeLength();
				muscle = this.mmm.muscle;
				exc = this.mmm.exc;
				y = [muscle.act,muscle.lM];
				dydt = muscle.calcDerivatives(y,exc,lMT);
				muscle.act = muscle.act + h*dydt(1);
				muscle.lM = muscle.lM + h*dydt(2);
				muscle.vM = dydt(2);
				% Compute auxiliary output
				muscle.calcAuxOutput(lMT);
			end
			if ~isempty(this.next)
				this.next.integrateHMM(h);
			end
		end

		%%
		function l = computeLength(this) %#ok<STOUT,MANU> 
			error('Unimplemented');
		end

		%%
		function setVa(this,va)
			this.vs = reshape(va(this.idxA),3,[]);
			if ~isempty(this.next)
				this.next.setVa(va);
			end
		end
		
		%%
		function [T,V] = computeEnergies(this,grav,T,V)
			if isempty(this.as)
				return
			end
			if nargin == 2
				T = 0;
				V = 0;
			end
			n = size(this.xs,2);
			for i = 1 : n
				mi = this.ms(i);
				xi = this.xs(:,i);
				vi = this.vs(:,i);
				V = V - mi*(grav'*xi);
				T = T + 0.5*mi*(vi'*vi);
			end
			if ~isempty(this.next)
				[T,V] = this.next.computeEnergies(grav,T,V);
			end
		end

		%%
		function storeNN(this) %#ok<MANU> 
		end

		%%
		function restoreNN(this) %#ok<MANU> 
		end

		%%
		function draw(this)
			if isempty(this.as)
				return
			end
			% Draw muscle path
			this.draw_();
			% Draw origin/insertion points
			x = [this.xOw,this.xIw];
			plot3(x(1,:),x(2,:),x(3,:),'ro','MarkerSize',4,'MarkerFaceColor','r');
			% Draw mass points
			x = this.xs;
			plot3(x(1,:),x(2,:),x(3,:),'ko','MarkerSize',4,'MarkerFaceColor','k');
			%plot3(x(1,:),x(2,:),x(3,:),'r-','LineWidth',2);
			if ~isempty(this.next)
				this.next.draw();
			end
		end
	end
	
	%%
	methods (Access = protected)
		
		%%
		function update_(this) %#ok<MANU>
		end

		%%
		function [Jam,dJam] = computeJacobian_(this,Jam,dJam,e) %#ok<INUSD,INUSL> 
		end

		%%
		function [JamNNr,JarNNr] = computeJacobianNN_(this,JamNNr,JarNNr) %#ok<INUSL> 
		end

		%%
		function points = getForcePointsHMM(this) %#ok<STOUT,MANU> 
			error('Unimplemented');
		end
		
		%%
		function draw_(this) %#ok<MANU>
		end

		%%
		function c = getColor(this)
			c = [1 0 0];
			if ~isempty(this.mmm)
				c(2) = 1 - this.mmm.muscle.act;
			end
		end
	end
end
