classdef MuscleII < redmax.Muscle
	%MuscleII A muscle going through path points
	%

	properties
		pathPoints % Origin, path points, and insertion
		xias       % Segment index and alpha for each mass point
		initLen    % Initial musculotendon length
	end

	%%
	methods
		%%
		function this = MuscleII(as,ms,pathPoints)
			bO = pathPoints(1).body;
			xO = pathPoints(1).xbody;
			bI = pathPoints(end).body;
			xI = pathPoints(end).xbody;
			this = this@redmax.Muscle(as,ms,bO,xO,bI,xI);
			this.pathPoints = pathPoints;
			this.xias = zeros(2,length(as));
			
			% Compute total length at rest
			this.initLen = this.computeLength();
		end

		%%
		function lTotal = computeLength(this)
			lTotal = 0;
			for i = 1 : length(this.pathPoints) - 1
				body0 = this.pathPoints(i).body;
				body1 = this.pathPoints(i+1).body;
				xbody0 = this.pathPoints(i).xbody(1:3);
				xbody1 = this.pathPoints(i+1).xbody(1:3);
				x0 = body0.E_wi(1:3,:)*[xbody0;1];
				x1 = body1.E_wi(1:3,:)*[xbody1;1];
				dx = x1 - x0;
				l = norm(dx);
				this.pathPoints(i).s = lTotal;
				lTotal = lTotal + l;
			end
		end
	end
		
	%%
	methods (Access = protected)
		%%
		function update_(this)
			% Update path Lagrangian positions and compute total length.
			% We'll use the current values of segment lengths to compute s
			% initially. Then, we'll divide these values by the total 
			% length, and then multiply by the initial total length to 
			% compute the correct s values.
			m = length(this.pathPoints);
			lTotal = 0;
			for i = 1 : m
				body = this.pathPoints(i).body;
				xbody = this.pathPoints(i).xbody(1:3);
				xworld = body.E_wi(1:3,:)*[xbody;1];
				this.pathPoints(i).x = xworld;
				this.pathPoints(i).s = 0;
				if i > 1
					x0 = this.pathPoints(i-1).x;
					x1 = this.pathPoints(i).x;
					dx = x1 - x0;
					l = norm(dx);
					this.pathPoints(i-1).dx = dx;
					this.pathPoints(i-1).l = l;
					lTotal = lTotal + l;
					this.pathPoints(i).s = lTotal;
				end
			end

			% Multiply s by the following ratio because we want s to be wrt to the
			% original total length, not the current total length.
			ratio = this.initLen/lTotal;
			this.pathPoints(1).s = 0;
			for i = 2 : m - 1
				this.pathPoints(i).s = ratio*this.pathPoints(i).s;
			end
			this.pathPoints(end).s = this.initLen;

			% Precompute per-segment deformation gradient
			for i = 1 : m - 1
				x0 = this.pathPoints(i).x;
				x1 = this.pathPoints(i+1).x;
				s0 = this.pathPoints(i).s;
				s1 = this.pathPoints(i+1).s;
				dx = x1 - x0;
				ds = s1 - s0;
				this.pathPoints(i).ds = ds;
				this.pathPoints(i).F = -dx/ds;
			end

			% Update mass positions
			% Convert percentage to length
			n = size(this.xs,2);
			ls = this.as*lTotal;
			x0 = this.pathPoints(1).x;
			x1 = this.pathPoints(2).x;
			l0 = 0;
			l1 = l0 + norm(x1 - x0);
			i = 1; % mass point index
			j = 3; % path point index (3 is next)
			while true
				if ls(i) < l1 + 1e-6
					% xi is in this segment
					a = (ls(i) - l0)/(l1 - l0);
					this.xs(:,i) = (1-a)*x0 + a*x1;
					this.xias(1,i) = j - 2; % segment index
					this.xias(2,i) = a; % alpha along this index
					i = i + 1;
					if i > n
						% Fix last segment index if necessary and switch alpha from 0 to 1.
						if this.xias(1,end) == m
							this.xias(1,end) = m - 1;
							this.xias(1,end) = 1.0;
						end
						break;
					end
				else
					% Go to the next segment
					x0 = x1;
					x1 = this.pathPoints(j).x;
					l0 = l1;
					l1 = l0 + norm(x1 - x0);
					j = j + 1;
				end
			end
		end

		%%
		function [Jam,dJam] = computeJacobian_(this,Jam,dJam,e) %#ok<INUSD> 
			% Jacobian from maximal coords to muscle points: Jar = Jaz*Jzx*Jxm
			idxA = this.idxA;
			na = length(this.as); % Number of mass points in this muscle only
			nx = length(this.pathPoints);
			nm = redmax.Scene.countM();

			% 1. Update path point velocities
			% 2. Jxm: vx = Jxm*vm
			%    rows = 3 times number of path points (x only)
			%    cols = number of maximal DOFs
			% 3. dJxm
			Jxm = zeros(3*nx,nm);
			dJxm = zeros(3*nx,nm);
			for i = 1 : nx
				% 1
				body = this.pathPoints(i).body;
				xbody = this.pathPoints(i).xbody(1:3);
				this.pathPoints(i).vx = body.E_wi(1:3,1:3)*se3.Gamma(xbody)*body.phi;
				this.pathPoints(i).vs = 0; % to be computed later via Jacobian
				% 2
				ii = 3*(i-1) + (1:3);
				jj = body.idxM;
				R = body.E_wi(1:3,1:3);
				G = se3.Gamma(xbody);
				Jxm(ii,jj) = R*G;
				% 3
				W = se3.brac(body.phi(1:3));
				dJxm(ii,jj) = R*W*G;
			end

			% Jsx: This must be computed fully before dJsx.
			L = zeros(nx-2,nx-2);
			S = zeros(nx-2,nx-1);
			X = zeros(nx-1,3*nx);
			for i = 1 : nx - 1
				% Fill X row
				dx = this.pathPoints(i).dx;
				l = this.pathPoints(i).l;
				dxbar = dx/l;
				jx0 = 3*(i-1) + (1:3);
				jx1 = jx0 + 3;
				X(i,[jx0,jx1]) = [-dxbar',dxbar'];
				% Fill S row
				if i > 1
					ds0 = this.pathPoints(i-1).ds;
					ds1 = this.pathPoints(i).ds;
					S(i-1,[i-1,i]) = [-ds1, ds0];
				end
				% Fill L row
				if i < nx - 1
					l0 = l;
					l1 = this.pathPoints(i+1).l;
					L(i,i) = l0 + l1;
					if i > 1
						L(i,i-1) = -l1;
					end
					if i < nx - 2
						L(i,i+1) = -l0;
					end
				end
			end
			Jsx = -(L\S)*X;

			% dJsx
			Ldot = zeros(nx-2,nx-2);
			Sdot = zeros(nx-2,nx-1);
			Xdot = zeros(nx-1,3*nx);
			vx = reshape([this.pathPoints.vx],[],1);
			vs = Jsx*vx;
			tmp = num2cell(vs',1);
			[this.pathPoints(2:end-1).vs] = tmp{:};
			for i = 1 : nx - 1
				vx0 = this.pathPoints(i).vx;
				vx1 = this.pathPoints(i+1).vx;
				vs0 = this.pathPoints(i).vs;
				vs1 = this.pathPoints(i+1).vs;
				this.pathPoints(i).dvx = vx1 - vx0;
				this.pathPoints(i).dvs = vs1 - vs0;
			end
			I = eye(3);
			for i = 1 : nx - 1
				% Fill Xdot row
				l = this.pathPoints(i).l;
				dx = this.pathPoints(i).dx;
				dvx = this.pathPoints(i).dvx;
				dxbar = dx/l;
				dvxbar = (I - dxbar*dxbar')*dvx/l;
				jx0 = 3*(i-1) + (1:3);
				jx1 = jx0 + 3;
				Xdot(i,[jx0,jx1]) = [-dvxbar',dvxbar'];
				% Fill Sdot row
				if i > 1
					dvs0 = this.pathPoints(i-1).dvs;
					dvs1 = this.pathPoints(i).dvs;
					Sdot(i-1,[i-1,i]) = [-dvs1, dvs0];
				end
				% Fill Ldot row
				if i < nx - 1
					l0 = l;
					dx0 = dx;
					dvx0 = dvx;
					l1 = this.pathPoints(i+1).l;
					dx1 = this.pathPoints(i+1).dx;
					dvx1 = this.pathPoints(i+1).dvx;
					l0dot = dx0'*dvx0/l0;
					l1dot = dx1'*dvx1/l1;
					Ldot(i,i) = l0dot + l1dot;
					if i > 1
						Ldot(i,i-1) = -l1dot;
					end
					if i < nx - 2
						Ldot(i,i+1) = -l0dot;
					end
				end
			end
			dJsx = -L\(Ldot*Jsx + Sdot*X + S*Xdot);

			% 1. Jzx: vz = Jzx*vx
			%    rows = 4 times number of path points (x and s)
			%    cols = 3 times number of path points (x only)
			% 2. dJzx = [0; dJsx]
			Jzx = zeros(4*nx,3*nx);
			dJzx = zeros(4*nx,3*nx);
			srows = [];
			for i = 1 : nx
				jx = 3*(i-1) + (1:3);
				izx = 4*(i-1) + (1:3);
				izs = izx(end) + 1;
				Jzx(izx,jx) = I;
				if 1 < i && i < nx
					srows(end+1) = izs; %#ok<AGROW> 
				end
			end
			Jzx(srows,:) = Jsx;
			dJzx(srows,:) = dJsx;

			% 1. Jaz: va = Jaz*vz
			%    rows = 3 times number of mass points
			%    cols = 4 times number of path points (x and s)
			% 2. dJaz
			Jaz = zeros(3*na,4*nx);
			dJaz = zeros(3*na,4*nx);
			for i = 1 : na
				% 1
				rows = 3*(i-1) + (1:3);
				j = this.xias(1,i);
				a = this.xias(2,i);
				F = this.pathPoints(j).F;
				jzx0 = 4*(j-1) + (1:3);
				jzs0 = jzx0(end) + 1;
				jzx1 = 4*(j) + (1:3);
				jzs1 = jzx1(end) + 1;
				Jaz(rows,jzx0) = (1 - a)*I;
				Jaz(rows,jzs0) = F*(1 - a);
				Jaz(rows,jzx1) = a*I;
				Jaz(rows,jzs1) = F*a;
				% 2
				ds = this.pathPoints(j).ds;
				dvx = this.pathPoints(j).dvx;
				dvs = this.pathPoints(j).dvs;
				vs0 = this.pathPoints(j).vs;
				vs1 = this.pathPoints(j+1).vs;
				dai = -((1 - a)/ds*vs0 + a/ds*vs1);
				dF = -(1/ds)*(dvx + F*dvs);
				dJaz(rows,jzx0) = -dai*I;
				dJaz(rows,jzs0) = -F*dai + dF*(1 - a);
				dJaz(rows,jzx1) = dai*I;
				dJaz(rows,jzs1) = F*dai + dF*a;
			end

			% Jam: va = Jam*vm
			Jam(idxA,:) = Jaz*Jzx*Jxm;
			% dJam
			dJam(idxA,:) = dJaz*Jzx*Jxm + Jaz*dJzx*Jxm + Jaz*Jzx*dJxm;
		end

		%%
		function points = getForcePointsHMM(this)
			b0 = this.pathPoints(1).body;
			x0 = b0.E_wi(1:3,:)*[this.pathPoints(1).xbody(1:3);1];
			points = [];
			for i = 2 : length(this.pathPoints)
				b1 = this.pathPoints(i).body;
				x1 = b1.E_wi(1:3,:)*[this.pathPoints(i).xbody(1:3);1];
				dx = x1 - x0;
				dx = dx/norm(dx);
				points(end+1).xw = x0; %#ok<AGROW> 
				points(end).dw = dx;
				points(end).body = b0;
				points(end+1).xw = x1; %#ok<AGROW> 
				points(end).dw = -dx;
				points(end).body = b1;
				b0 = b1;
				x0 = x1;
			end
		end
		
		%%
		function draw_(this)
			% Draw muscle path
			x = [this.pathPoints.x];
			c = this.getColor();
			plot3(x(1,:),x(2,:),x(3,:),'-','Color',c,'LineWidth',2);
			% Draw path points
			%plot3(x(1,2:end-1),x(2,2:end-1),x(3,2:end-1),'sc','MarkerSize',6,'LineWidth',1);
		end
	end
end
