classdef MuscleIII < redmax.MuscleNN
	%MuscleIII_debug3 A muscle wrapping around a wrap object.
	%
	% This class simulates the use of a NN. To actually use a NN, replace
	% the code in xwrap() and Jwrap() functions.
	%
	% This class uses a (simulated) NN that takes as input the origin and
	% insertion points, rather than the joint angle. This means that the
	% same NN can be used for uni-, bi-, and multi-articular muscles as
	% well as for different types of joints.
	
	properties
		wrap   % Wrapping object
	end

	%%
	methods
		%%
		function this = MuscleIII(as,ms,bO,xO,bI,xI,wrap,NNfile)
			this = this@redmax.MuscleNN(as,ms,bO,xO,bI,xI,NNfile);
			this.wrap = wrap;
		end
	end
	
	%%
	methods (Access = protected)
		%%
		function update_(this)
			n = length(this.as);
			[E,R,GO,GI] = computeTransforms(this);
			
			% Transform origin and insertion to wrapping surface (S)
			xO_S = E.S_O(1:3,:)*[this.xO;1];
			xI_S = E.S_I(1:3,:)*[this.xI;1];
			
			% Collect input and evaluate NN
			x = zeros(8,n);
			x(1,:) = this.as;
			x(2:4,:) = repmat(xO_S,1,n);
			x(5:7,:) = repmat(xI_S,1,n);
			x(8,:) = repmat(this.wrap.radius,1,n);
			[this.xsNN,this.JsNN] = redmax.MuscleIII.evaluateReverse(x,this.NN);

			% Finite differencing for the dot terms
			phi_B = this.wrap.body.phi;
			phi_O = this.bO.phi;
			phi_I = this.bI.phi;
			vO_S = R.S_W*(R.W_O*GO.O*phi_O - R.W_B*GO.B*phi_B);
			vI_S = R.S_W*(R.W_I*GI.I*phi_I - R.W_B*GI.B*phi_B);
			e = 1e-4;
			xO_S_ = xO_S + e*vO_S;
			xI_S_ = xI_S + e*vI_S;
			x_ = x;
			x_(2:4,:) = repmat(xO_S_,1,n);
			x_(5:7,:) = repmat(xI_S_,1,n);
			[xsNN_,JsNN_] = redmax.MuscleIII.evaluateReverse(x_,this.NN);
			this.dxsNN = (xsNN_ - this.xsNN)/e;
			this.dJsNN = (JsNN_ - this.JsNN)/e;

			% Transform from S to world
			this.xs = E.W_S(1:3,:)*[this.xsNN;ones(1,n)];
		end

		%%
		function [Jam,dJam] = computeJacobian_(this,Jam,dJam,e) %#ok<INUSD> 
			% Notation:
			% S = wrapping surface
			% B = body that contains S
			% O = body that contains the origin
			% I = body that contains the insertion
			idxMB = this.wrap.body.idxM;
			idxMO = this.bO.idxM;
			idxMI = this.bI.idxM;

			[E,R,GO,GI] = computeTransforms(this);

			% Derivatives
			phi_B = this.wrap.body.phi;
			phi_O = this.bO.phi;
			phi_I = this.bI.phi;
			phibrac_B = se3.brac(phi_B);
			phibrac_O = se3.brac(phi_O);
			phibrac_I = se3.brac(phi_I);
			dR_W_B = R.W_B*se3.brac(phi_B(1:3));
			dR_W_O = R.W_O*se3.brac(phi_O(1:3));
			dR_W_I = R.W_I*se3.brac(phi_I(1:3));
			dR_W_S = dR_W_B*R.B_S;
			dR_S_W = dR_W_S';
			dGO_B = [-se3.brac((E.B_O*phibrac_O - phibrac_B*E.B_O)*[this.xO;1]), zeros(3)];
			dGI_B = [-se3.brac((E.B_I*phibrac_I - phibrac_B*E.B_I)*[this.xI;1]), zeros(3)];

			n = size(this.xs,2);
			for i = 1 : n
				idxAi = this.idxA(3*(i-1)+(1:3));
				x_S = this.xsNN(:,i);
				J_S = this.JsNN(:,i);
				dx_S = this.dxsNN(:,i);
				dJ_S = this.dJsNN(:,i);
				% Surface base term
				x_B = E.B_S*[x_S; 1];
				G_B = se3.Gamma(x_B);
				dG_B = [-se3.brac(R.B_S*dx_S), zeros(3)];
				Jam(idxAi,idxMB) = R.W_B*G_B;
				dJam(idxAi,idxMB) = dR_W_B*G_B + R.W_B*dG_B;
				% Origin relative term
				JO_S = reshape(J_S(1:9),3,3);
				dJO_S = reshape(dJ_S(1:9),3,3);
				JO_W = R.W_S*JO_S*R.S_W;
				dJO_W = dR_W_S*JO_S*R.S_W + R.W_S*dJO_S*R.S_W + R.W_S*JO_S*dR_S_W;
				Jam(idxAi,idxMO) = Jam(idxAi,idxMO) + JO_W*R.W_O*GO.O;
				Jam(idxAi,idxMB) = Jam(idxAi,idxMB) - JO_W*R.W_B*GO.B;
				dJam(idxAi,idxMO) = dJam(idxAi,idxMO) + dJO_W*R.W_O*GO.O + JO_W*dR_W_O*GO.O; % dGO_O = 0;
				dJam(idxAi,idxMB) = dJam(idxAi,idxMB) - dJO_W*R.W_B*GO.B - JO_W*dR_W_B*GO.B - JO_W*R.W_B*dGO_B;
				% Insertion relative term
				JI_S = reshape(J_S(10:18),3,3);
				dJI_S = reshape(dJ_S(10:18),3,3);
				JI_W = R.W_S*JI_S*R.S_W;
				dJI_W = dR_W_S*JI_S*R.S_W + R.W_S*dJI_S*R.S_W + R.W_S*JI_S*dR_S_W;
				Jam(idxAi,idxMI) = Jam(idxAi,idxMI) + JI_W*R.W_I*GI.I;
				Jam(idxAi,idxMB) = Jam(idxAi,idxMB) - JI_W*R.W_B*GI.B;
				dJam(idxAi,idxMI) = dJam(idxAi,idxMI) + dJI_W*R.W_I*GI.I + JI_W*dR_W_I*GI.I; % dGI_I = 0;
				dJam(idxAi,idxMB) = dJam(idxAi,idxMB) - dJI_W*R.W_B*GI.B - JI_W*dR_W_B*GI.B - JI_W*R.W_B*dGI_B;
			end
		end
		
		%%
		function draw_(this)
			% Draw wrapping surface
			this.wrap.draw();
			% Instead of drawing the path, we will simply connect the mass points.
			% This will not work unless the distribution is [0,1].
			plot3(this.xs(1,:),this.xs(2,:),this.xs(3,:),'r-','LineWidth',2);
		end

		%%
		function [E,R,GO,GI] = computeTransforms(this)
			% Transforms between wrapping surface (S) and its base body (B)
			E.B_S = this.wrap.E;
			E.W_B = this.wrap.body.E_wi;
			E.B_W = se3.inv(E.W_B);
			E.W_S = E.W_B*E.B_S;
			E.S_W = se3.inv(E.W_S);
			R.B_S = E.B_S(1:3,1:3);
			R.W_B = E.W_B(1:3,1:3);
			R.W_S = E.W_S(1:3,1:3);
			R.S_W = R.W_S';

			% Transforms between origin body (O) and wrapping surface (S)
			E.W_O = this.bO.E_wi;
			R.W_O = E.W_O(1:3,1:3);
			E.B_O = E.B_W*E.W_O;
			E.S_O = E.S_W*E.W_O;
			xO_O = this.xO;
			xO_B = E.B_O*[xO_O; 1];
			GO.O = se3.Gamma(xO_O);
			GO.B = se3.Gamma(xO_B);

			% Transforms between insertion body (I) and wrapping surface (S)
			E.W_I = this.bI.E_wi;
			R.W_I = E.W_I(1:3,1:3);
			E.B_I = E.B_W*E.W_I;
			E.S_I = E.S_W*E.W_I;
			xI_I = this.xI;
			xI_B = E.B_I*[xI_I; 1];
			GI.I = se3.Gamma(xI_I);
			GI.B = se3.Gamma(xI_B);
		end
	end

	%%
	methods (Static)
		%%
		function [z,g] = evaluateReverse(x,NN)
			% g is the gradient wrt the ith input
			W1 = NN.W1;
			b1 = NN.b1;
			W2 = NN.W2;
			b2 = NN.b2;
			W3 = NN.W3;
			b3 = NN.b3;
			W4 = NN.W4;
			b4 = NN.b4;
			W5 = NN.W5;
			b5 = NN.b5;
			W6 = NN.W6;
			b6 = NN.b6;
			W7 = NN.W7;
			b7 = NN.b7;
			W8 = NN.W8;
			b8 = NN.b8;

			y1 = W1*x + b1;
			z1 = tanh(y1);
			y2 = W2*z1 + b2;
			z2 = tanh(y2);
			y3 = W3*z2 + b3;
			z3 = tanh(y3);
			y4 = W4*z3 + b4;
			z4 = tanh(y4);
			y5 = W5*z4 + b5;
			z5 = tanh(y5);
			y6 = W6*z5 + b6;
			z6 = tanh(y6);
			y7 = W7*z6 + b7;
			z7 = tanh(y7);
			y8 = W8*z7 + b8;
			z = y8;

			if nargout == 1
				return;
			end

			% Compute gradient in reverse mode

			nx = size(x,2);
			D1 = (1 - z1.^2)';
			D2 = (1 - z2.^2)';
			D3 = (1 - z3.^2)';
			D4 = (1 - z4.^2)';
			D5 = (1 - z5.^2)';
			D6 = (1 - z6.^2)';
			D7 = (1 - z7.^2)';

			ix = 1 : nx;
			iy = nx+1 : 2*nx;
			iz = 2*nx+1 : 3*nx;
			tmp7(ix,:) = D7 .* W8(1,:);
			tmp7(iy,:) = D7 .* W8(2,:);
			tmp7(iz,:) = D7 .* W8(3,:);

			tmp6 = tmp7 * W7;
			tmp6 = repmat(D6,3,1) .* tmp6;
			tmp5 = tmp6 * W6;
			tmp5 = repmat(D5,3,1) .* tmp5;
			tmp4 = tmp5 * W5;
			tmp4 = repmat(D4,3,1) .* tmp4;
			tmp3 = tmp4 * W4;
			tmp3 = repmat(D3,3,1) .* tmp3;
			tmp2 = tmp3 * W3;
			tmp2 = repmat(D2,3,1) .* tmp2;
			tmp1 = tmp2 * W2;
			tmp1 = repmat(D1,3,1) .* tmp1;

			g = zeros(18,nx);
			for i = 1:6
				W1i = W1(:,1+i);
				g(3*i-2,:) = tmp1(ix,:) * W1i;
				g(3*i-1,:) = tmp1(iy,:) * W1i;
				g(3*i,:) = tmp1(iz,:) * W1i;
			end
		end
	end
end
