classdef Scene < handle
	%Scene Container class for things to be simulated
	
	properties
		name % scene name
		integrator % integrator type: 1=EULER, 2=BDF1, 3=SDIRK2
		bodies % list of bodies
		joints % list of joints
		muscles % list of muscles
		qInit % initial positions
		qdotInit % initial velocities
		tEnd % end time
		t % current time
		h % time step
		k % current step
		nsteps % number of steps to take
		drawHz % refresh rate (0 for no draw)
		grav % gravity
		plotH % whether to plot the energy at the end
		T0 % initial kinetic energy
		V0 % initial potential energy
		history % step history: does not include the initial state
		task % task for the adjoint method
		Vexpected % expected potential energy for testing
		wview % initial viewing angle
		waxis % initial axis
		wgrid % grid on/off
		walpha % display alpha
		winterp % interpolate frames for slow drawing
	end
	
	methods
		function this = Scene()
			% Default values
			redmax.Scene.clear();
			this.name = '';
			this.integrator = 1; % EULER
			this.bodies = {};
			this.joints = {};
			this.muscles = {};
			this.qInit = [];
			this.qdotInit = [];
			this.tEnd = 1.0;
			this.t = 0;
			this.h = 1e-2;
			this.k = 0;
			this.nsteps = 1;
			this.drawHz = 10;
			this.grav = [0 0 -980]';
			this.plotH = false;
			this.T0 = 0;
			this.V0 = 0;
			this.history = [];
			this.task = [];
			this.Vexpected = zeros(3,1);
			this.waxis = [];
			this.wview = [];
			this.wgrid = true;
			this.walpha = 0.2;
			this.winterp = 1;
			if usejava('jvm')
				colormap('default'); % Restore colormap mode
			end
		end
		
		function initJoints(this)
			% Joints
			njoints = length(this.joints);
			jroot = this.joints{1}; % Assumes joints{1} is root
			order = jroot.getTraversalOrder();
			this.joints = this.joints(order); % Reorders joints
			for i = njoints : -1 : 1
				% Leaf-to-root ordering is better for reducing matrix fill
				this.joints{i}.countDofs(); % This will also call Body.countDofs()
			end
			for i = 1 : njoints
				if i < njoints
					this.joints{i}.next = this.joints{i+1};
				end
				if i > 1
					this.joints{i}.prev = this.joints{i-1};
				end
			end
			for i = 1 : njoints
				this.joints{i}.computeInertia();
			end
			jroot.update();
			[this.qInit,this.qdotInit] = jroot.getQ();

			% Bodies
			nbodies = length(this.bodies);
			this.bodies = this.bodies(order); % same ordering as joints
			for i = 1 : nbodies
				if i < nbodies
					this.bodies{i}.next = this.bodies{i+1}; %#ok<*SAGROW>
				end
			end
		end
		
		%%
		function initMuscles(this)
			% Muscles
			nmuscles = length(this.muscles);
			for i = 1 : nmuscles
				this.muscles{i}.countDofs();
				this.muscles{i}.update();
				if i < nmuscles
					this.muscles{i}.next = this.muscles{i+1};
				end
			end
			if nmuscles == 0
				this.muscles{1} = redmax.MuscleNull();
			end
		end

		%%
		function init(this)
			this.nsteps = ceil(this.tEnd/this.h);
			[this.qInit,this.qdotInit] = this.joints{1}.getQ();
			this.reset();
		end
		
		%%
		function reset(this)
			jroot = this.joints{1};
			mroot = this.muscles{1};
			jroot.setQ(this.qInit,this.qdotInit);
			jroot.setQ0(this.qInit);
			this.t = 0;
			this.k = 0;
			[this.T0,this.V0] = jroot.computeEnergies(this.grav);
			[this.T0,this.V0] = mroot.computeEnergies(this.grav,this.T0,this.V0);
			if ~isempty(this.task)
				this.task.P = 0;
			end
		end
		
		%%
		function update(this,q,qdot,qadot)
			% Updates position and velocity information
			this.joints{1}.setQ(q,qdot);
			this.joints{1}.update();
			this.muscles{1}.update();
			if nargin == 4
				% If provided, set muscle mass point velocities
				this.muscles{1}.setVa(qadot);
			end
		end
		
		%%
		function updateQuick(this,q)
			% Only updates position information
			this.joints{1}.setQ(q);
			this.joints{1}.updateQuick();
			this.muscles{1}.update(); % quick version?
		end
		
		%%
		function saveHistory(this,history)
			[q,qdot] = this.joints{1}.getQ();
			this.history(this.k).q = q;
			this.history(this.k).qdot = qdot;
			if ~isempty(this.task) && nargin > 1
				% Used by the adjoint method
				this.history(this.k).H = history.H;
				this.history(this.k).M = history.M;
				this.history(this.k).D = history.D;
				this.history(this.k).Jmr = history.Jmr;
				this.task.calcStep();
			end
			[T,V] = this.joints{1}.computeEnergies(this.grav);
			[T,V] = this.muscles{1}.computeEnergies(this.grav,T,V);
			this.history(this.k).T = T;
			this.history(this.k).V = V;
			this.history(this.k).t = this.t;
		end
		
		%%
		function plotEnergies(this)
			T = [this.T0,this.history.T];
			V = [this.V0,this.history.V];
			t = [0,this.history.t]; %#ok<PROP> 
			V = V - V(1);
			H = T + V;
			if this.Vexpected(this.integrator) ~= 0
				dV = V(end) - this.Vexpected(this.integrator);
				if abs(dV) > 1e-2
					fprintf(2,'### FAIL: %.16e ###\n',V(end));
				else
					fprintf(1,'### PASS ###\n');
				end
			end
			if this.plotH
				clf;
				plot(t,T,'.-',t,V,'.-',t,H,'.-'); %#ok<PROP> 
				a = axis();
				a(2) = t(end); %#ok<PROP> 
				axis(a);
				grid on;
				legend('T','V','H');
				xlabel('Time');
				ylabel('Energy');
			end
		end
		
		%%
		function draw(this)
			if this.drawHz == 0
				return;
			end
			if this.t == 0
				clf;
				xlabel('X');
				ylabel('Y');
				zlabel('Z');
				axis equal;
				if ~isempty(this.waxis)
					axis(this.waxis);
				end
				ax = gca;
				ax.Clipping = 'off';
				if this.wgrid
					grid on;
				else
					grid off;
					set(gca,'XColor', 'none','YColor','none','ZColor','none');
				end
				if ~isempty(this.wview)
					view(this.wview);
				else
					view(3);
				end
			end
			if floor(this.t*this.drawHz) > floor((this.t-this.h)*this.drawHz)
				% Hack for drawing N interpolated frames
				q0 = this.joints{1}.getQ0();
				q = this.joints{1}.getQ();
				for i = 1 : this.winterp
					s = i/this.winterp;
					qs = (1 - s)*q0 + s*q;
					this.updateQuick(qs);
					cla;
					hold on;
					this.bodies{1}.draw();
					this.joints{1}.draw();
					this.muscles{1}.draw();
					if ~isempty(this.task)
						this.task.draw();
					end
					alpha(this.walpha);
					if this.wgrid
						title(sprintf('t = %.4f', this.t));
					end
					drawnow;
				end
				this.updateQuick(q);
			end
		end
	end
	
	%%
	methods (Static)
		%%
		function clear()
			global nm nr na countCM CM; %#ok<GVMIS> 
			nm = 0;
			nr = 0;
			na = 0;
			countCM = 1;
			if ~usejava('jvm')
				CM = zeros(1,3);
			else
				CM = colormap('lines');
			end
		end
		
		%%
		function out = countM(data)
			global nm; %#ok<GVMIS> 
			if nargin
				nm = data;
			end
			if isempty(nm)
				out = 0;
			else
				out = nm;
			end
		end
		
		%%
		function out = countR(data)
			global nr; %#ok<GVMIS> 
			if nargin
				nr = data;
			end
			if isempty(nr)
				out = 0;
			else
				out = nr;
			end
		end
		
		%%
		function out = countA(data)
			global na; %#ok<GVMIS> 
			if nargin
				na = data;
			end
			if isempty(na)
				out = 0;
			else
				out = na;
			end
		end
		
		%%
		function printError(name,v0,v1)
			thresh = 1e-2;
			s = size(v0);
			if length(s) < 3
				s = [s,1];
			end
			for k = 1 : s(3)
				abserr = norm(v1(:,:,k) - v0(:,:,k));
				relerr = abserr/norm(v1(:,:,k)); % assume v1 is FD
				err = min(abserr,relerr);
				if s(3) == 1
					namek = name;
				else
					namek = sprintf('%s%d',name,k);
				end
				str = pad(namek,16);
				if err < thresh
					fprintf('%s%e\n',str,err);
				else
					fprintf(2,'%s%e\n',str,err);
				end
			end
		end
	end
end
