classdef WrapCylinder < redmax.Wrap
	%WrapCylinder A cylinder wrapping object
	%   

	properties
		radius
		height   % only for display
		quadrant % not used by NN
	end

	methods
		%%
		function this = WrapCylinder(body,E,radius,height,quadrant)
			this = this@redmax.Wrap(body,E);
			this.radius = radius;
			this.height = height;
			this.quadrant = quadrant;
		end

		%%
		function x = evalSurf(this,z,a)
			% Evaluates the surface point in surface space. To go to world
			% space, multiply by:
			%    this.body.E_wi*this.E
			r = this.radius;
			x = [r*cos(a); r*sin(a); z];
		end

		%%
		function [zO,aO,zI,aI,l] = computeTanPts(this,xOo,xIo,first)
			r = this.radius;
			xOo = xOo(1:3);
			xIo = xIo(1:3);

			% Optimize for aO and aI
			%           [r*cos(a)]
			%    x(a) = [r*sin(a)]
			%           [    0   ]
			% Find a such that x(a)'*(x(a) - xO) = 0. The tangent angle
			% does not depend on the z coordinate.
			if true
				% Maybe better to do this all the time rather than just the first time?
				daMin = inf;
				for k = 1 : 4
					aInit = (k-1)/3*pi*2;
					aOk = redmax.Wrap.newton(@(a)redmax.WrapCylinder.tangential(a,r,xOo),aInit);
					aIk = redmax.Wrap.newton(@(a)redmax.WrapCylinder.tangential(a,r,xIo),aInit);
					if strcmp(this.quadrant,'+x')
						aMid = 0;
					elseif strcmp(this.quadrant,'-x')
						aMid = pi;
					elseif strcmp(this.quadrant,'+y')
						aMid = pi/2;
					else
						aMid = -pi/2;
					end
					daO = abs(aOk - aMid);
					daI = abs(aIk - aMid);
					da = daO + daI;
					if da < daMin
						daMin = da;
						aO = aOk;
						aI = aIk;
					end
				end
			else
				if strcmp(this.quadrant,'+x')
					aInit = 0;
				elseif strcmp(this.quadrant,'-x')
					aInit = pi;
				elseif strcmp(this.quadrant,'+y')
					aInit = pi/2;
				else
					aInit = -pi/2;
				end
				aO = redmax.Wrap.newton(@(a)redmax.WrapCylinder.tangential(a,r,xOo),aInit);
				aI = redmax.Wrap.newton(@(a)redmax.WrapCylinder.tangential(a,r,xIo),aInit);
			end
			if abs(aO - aI) < 1e-2
				%fprintf('About to lift off!\n');
			end
			
			% Optimize for zO and zI
			zInit = [0 0]';
			z = redmax.Wrap.newton(@(z)redmax.WrapCylinder.totalLength(z,aO,aI,r,xOo,xIo),zInit);
			zO = z(1);
			zI = z(2);
			l = redmax.WrapCylinder.totalLength(z,aO,aI,r,xOo,xIo);
		end

		%%
		function draw(this)
			E_wo = this.body.E_wi*this.E;
			s = this.getAxisSize();
			if s > 0
				se3.drawAxis(E_wo,s,3);
			end
			n = 16;
			r = this.radius;
			h = this.height;
			%{
			if strcmp(this.quadrant,'+x')
				thetaRange = [-pi/2,pi/2];
			elseif strcmp(this.quadrant,'-x')
				thetaRange = [pi/2,3*pi/2];
			elseif strcmp(this.quadrant,'+y')
				thetaRange = [0,pi];
			elseif strcmp(this.quadrant,'-y')
				thetaRange = [pi,2*pi];
			else
				thetaRange = [0,2*pi];
			end
			%}
			thetaRange = [0,2*pi]; % just draw the full cylinder
			[X,Y,Z] = se3.surfCylinder(E_wo,r,h,n,thetaRange);
			surf(X,Y,Z,'FaceColor',this.getColor(),'EdgeColor','none');
			% Draw outline
			x = zeros(3,0);
			thetas = linspace(thetaRange(1),thetaRange(2),n);
			for k = 1 : n
				theta = thetas(k);
				x(:,end+1) = E_wo(1:3,:)*[r*cos(theta),r*sin(theta),h/2,1]'; %#ok<*AGROW> 
			end
			x(:,end+1) = nan(3,1);
			for k = 1 : n
				theta = thetas(k);
				x(:,end+1) = E_wo(1:3,:)*[r*cos(theta),r*sin(theta),-h/2,1]';
			end
			n = 4;
			for k = 1 : n
				theta = (k-1)/n*2*pi;
				x(:,end+1) = nan(3,1);
				x(:,end+1) = E_wo(1:3,:)*[r*cos(theta),r*sin(theta),h/2,1]';
				x(:,end+1) = E_wo(1:3,:)*[r*cos(theta),r*sin(theta),-h/2,1]';
			end
			plot3(x(1,:),x(2,:),x(3,:),'k-');
		end

	end

	methods (Static)
		%%
		function [xa,dxa] = evalCircle(a,r)
			xa = [r*cos(a); r*sin(a); 0];
			if nargout == 2
				dxa = [-r*sin(a); r*cos(a); 0];
			end
		end
		
		%%
		function [f,g,H] = tangential(a,r,xt)
			[xa,dxa] = redmax.WrapCylinder.evalCircle(a,r);
			g = xa'*(xa - xt);
			H = (2*xa - xt)'*dxa;
			f = 0; % unused, since we're solving for g=0.
		end

		%%
		function [f,g,H] = totalLength(z,aO,aI,r,xOo,xIo)
			% Helical length:
			% Given two cylindrical coordinate points p1 and p2:
			%    p1
			%    |\
			%    | \
			%    |  \
			%    |___\
			%        p2
			% the geodesic length is sqrt(dz^2 + (r*da)^2), where dz is
			% the difference in z and da is the difference in a. To compute
			% the total length of the muscle, add the two straight line
			% segments from the origin to the wrapping object and then to
			% the insertion.
			%{
			syms r aO aI real
			syms z [2 1] real
			syms xOo xIo [3 1] real
			xOz = evalXz(z(1),aO,r);
			xIz = evalXz(z(2),aI,r);
			dxO = xOz - xOo; % origin to wrap object
			dxI = xIo - xIz; % wrap object to insertion
			lO = sqrt(dxO'*dxO);
			lI = sqrt(dxI'*dxI);
			dz = z(2) - z(1);
			da = aI - aO;
			lW = sqrt(dz^2 + (r*da)^2);
			f = lO + lW + lI;
			g = gradient(f,z);
			H = jacobian(g,z);
			matlabFunction(f,g,H,'File','wrap','Optimize',true);
			function xz = evalXz(z,a,r)
				xz = [r*cos(a); r*sin(a); z];
			end
			%}
			z1 = z(1);
			z2 = z(2);
			xOo1 = xOo(1);
			xOo2 = xOo(2);
			xOo3 = xOo(3);
			xIo1 = xIo(1);
			xIo2 = xIo(2);
			xIo3 = xIo(3);
			%
			t2 = cos(aI);
			t3 = cos(aO);
			t4 = sin(aI);
			t5 = sin(aO);
			t6 = r.^2;
			t7 = xIo3.*2.0;
			t8 = xOo3.*2.0;
			t9 = z1.*2.0;
			t10 = z2.*2.0;
			t15 = -aO;
			t16 = -z1;
			t18 = -z2;
			t11 = r.*t2;
			t12 = r.*t3;
			t13 = r.*t4;
			t14 = r.*t5;
			t17 = -t9;
			t19 = -t10;
			t20 = aI+t15;
			t25 = t18+xIo3;
			t26 = t16+xOo3;
			t27 = t18+z1;
			t28 = t7+t19;
			t29 = t25.^2;
			t30 = t8+t17;
			t31 = t26.^2;
			t32 = t9+t19;
			t33 = t27.^2;
			t38 = t20.^2;
			t40 = (t11-xIo1).^2;
			t41 = (t12-xOo1).^2;
			t42 = (t13-xIo2).^2;
			t43 = (t14-xOo2).^2;
			t39 = t32.^2;
			t44 = t6.*t38;
			t48 = t29+t40+t42;
			t49 = t31+t41+t43;
			t45 = t33+t44;
			t51 = 1.0./sqrt(t48);
			t52 = 1.0./sqrt(t49);
			t46 = 1.0./sqrt(t45);
			f = 1.0./t46+1.0./t51+1.0./t52;
			if nargout > 1
				t47 = t46.^3;
				t50 = -t46;
				t53 = (t32.*t46)./2.0;
				g = [t53-(t30.*t52)./2.0;-t53-(t28.*t51)./2.0];
			end
			if nargout > 2
				t54 = (t39.*t47)./4.0;
				t55 = -t54;
				t56 = t50+t54;
				H = reshape([t46+t52+t55-(t30.^2.*t52.^3)./4.0,t56,t56,t46+t51+t55-(t28.^2.*t51.^3)./4.0],[2,2]);
			end
		end
	end
end
