#include "WrapSphere.hpp"

void WrapSphere::compute()
{
    Eigen::Vector3f OS = this->point_S - this->point_O;
    OS = OS / OS.norm();
    Eigen::Vector3f OP = this->point_P - this->point_O;
    OP = OP / OP.norm();
    Eigen::Vector3f N = OP.cross(OS);
    if (N.norm() != 0) {
        N = N / N.norm();
    }
    
    this->M << OS.transpose(), N.cross(OS).transpose(), N.transpose();

    Eigen::Vector3f p = this->M * (this->point_P - this->point_O);
    Eigen::Vector3f s = this->M * (this->point_S - this->point_O);

    double denom_q = p(0)*p(0) + p(1)*p(1);
    double denom_t = s(0)*s(0) + s(1)*s(1);
    double R = this->radius;

    this->status = Status::wrap;

    if ((denom_q - R*R < 0.0) || (denom_t - R*R < 0.0))
    {
        this->status = Status::inside_radius;
    }
        
    double root_q = sqrt(denom_q - R*R);
    double root_t = sqrt(denom_t - R*R);

    Eigen::Vector3f q(0.0f, 0.0f, 0.0f);
    Eigen::Vector3f t(0.0f, 0.0f, 0.0f);
    q(0) = (p(0) * R*R - R * p(1) * root_q) / denom_q;
    q(1) = (p(1) * R*R + R * p(0) * root_q) / denom_q;
    t(0) = (s(0) * R*R + R * s(1) * root_t) / denom_t;
    t(1) = (s(1) * R*R - R * s(0) * root_t) / denom_t;

    if (R * (q(0) * t(1) - q(1) * t(0)) < 0.0)
    {
        this->status = Status::no_wrap;
    }
    
    this->point_q = q;
    this->point_t = t;
    
    Eigen::Vector3f Q = this->M.transpose() * q + this->point_O;
    Eigen::Vector3f T = this->M.transpose() * t + this->point_O;
    this->point_Q = Q;
    this->point_T = T;

    if (this->status == Status::inside_radius || this->status == Status::no_wrap) {
        this->total_length = (this->point_P - this->point_S).norm();
        this->path_length = 0.0f;
    }

    if (this->status == Status::wrap) {
        this->path_length = R * acos(1.0 - 0.5 *
                                     ((q(0) - t(0)) * (q(0) - t(0))
                                      + (q(1) - t(1)) * (q(1) - t(1))) / (R*R));
        if (this->path_length < 0.0) {
            this->path_length = -this->path_length;
        }

        float len0 = (this->point_P - this->point_Q).norm();
        float len1 = this->getLength();
        float len2 = (this->point_S - this->point_T).norm();
        this->len0 = len0;
        this->len2 = len2;
        assert(len0 >= 0 && len1 >= 0 && len2 >= 0);
        float len = len0 + len1 + len2;
        this->total_length = len;
    }
}

Eigen::Vector3f WrapSphere::getPoints(float alpha)
{
    double theta_q = atan(this->point_q(1) / this->point_q(0));
    if (this->point_q(0) < 0.0f)
        theta_q += PI;
    
    double theta_t = atan(this->point_t(1) / this->point_t(0));
    if (this->point_t(0) < 0.0f)
        theta_t += PI;
    
    Eigen::Vector3f point;
    double theta_s, theta_e;
    
    if (theta_q < theta_t)
    {
        theta_s = theta_q; theta_e = theta_t;
    }
    else
    {
        theta_s = theta_t; theta_e = theta_q;
    }

    if (theta_e - theta_s > theta_s + 2*PI - theta_e)
    {
        double tmp = theta_s; theta_s = theta_e; theta_e = tmp + 2*PI;
    }

    if (this->status == Status::wrap) {

        float len = this->getTotalLength();
        float len0 = this->len0;
        float len2 = this->len2;
        float len1 = this->getLength();

        if (alpha <= len0 / len) {
            float s = len * alpha / len0;
            point = (1 - s) * this->point_P + s * this->point_Q;
        }
        else if (alpha >= (len0 + len1) / len) {
            float s = (alpha * len - len0 - len1) / len2;
            point = (1 - s) * this->point_T + s * this->point_S;
        }
        else
        {
            float s = (alpha * len - len0) / len1;
            float v = (1 - s) * theta_e + s * theta_s;
            point = this->M.transpose() *
                Eigen::Vector3f(this->radius * cos(v), this->radius * sin(v), 0.0f) +
                this->point_O;
        }
    }

    if (this->status == Status::inside_radius || this->status == Status::no_wrap) {

        float s = alpha;
        point = (1 - s) * this->point_P + s * this->point_S;
    }

    return point;
}
