import GPUtil
import torch
import os
import torch.nn as nn
import torch.onnx
from torch.optim import lr_scheduler
from torch.utils.data import DataLoader
from torch.utils.data import Dataset
from torch.utils.tensorboard import SummaryWriter
import numpy as np
import argparse

class CustomDataset(Dataset):
	def __init__(self, x_tensor, y_tensor, device):
		self.x = x_tensor
		self.y = y_tensor
		self.device = device

	def __getitem__(self, index):
		data = self.x[index]
		targets = self.y[index]
		return (data, targets)

	def __len__(self):
		return len(self.x)

class Feedforward(torch.nn.Module):
	def __init__(self, input_size=8, hidden_size=256, output_size=3, n_layers=6, nonlinear=nn.Tanh):
		super(Feedforward, self).__init__()
		self.input_size = input_size
		self.hidden_size = hidden_size
		self.output_size = output_size
		self.n_layers = n_layers
		self.nonlinear = nonlinear
		self.fc1 = torch.nn.Linear(self.input_size, self.hidden_size)
		self.relu = torch.nn.ReLU()
		self.fc2 = torch.nn.Linear(self.hidden_size, self.output_size)

		layers = [nn.Linear(self.input_size, self.hidden_size)]
		for i in range(self.n_layers):
			layers.append(nonlinear())
			layers.append(nn.Linear(hidden_size, hidden_size))
		layers.append(nonlinear())
		layers.append(nn.Linear(hidden_size, output_size))
		self.layers = nn.Sequential(*layers)

	def forward(self, x):
		x = self.layers(x)
		return x

def save(net1, save_path, epoch):
	net1 = net1.eval()
	torch.save(net1, os.path.join(save_path, "net_wrap_" + str(epoch) + ".pkl"))
	torch.save(net1.state_dict(), os.path.join(save_path, "net_params_wrap_" + str(epoch) + ".pkl"))

def make_train_step(model, loss_fn, optimizer):
	def train_step(x, y):
		model.train()
		yhat = model(x)
		loss = loss_fn(y, yhat)
		loss.backward()
		optimizer.step()
		optimizer.zero_grad()
		return loss.item()
	return train_step

def train(net1, save_path):
	optimizer = torch.optim.Adam(net1.parameters(), lr=learning_rate)
	scheduler = lr_scheduler.StepLR(optimizer, step_size=step_size_, gamma=gamma_)
	loss_func = torch.nn.MSELoss(reduction='mean')

	loss_values = []
	valoss_values = []
			
	x_train_offline = np.loadtxt(datasets_path + 'x_train_offline.txt',dtype=float)
	y_train_offline = np.loadtxt(datasets_path + 'y_train_offline.txt',dtype=float)
	x_test_offline = np.loadtxt(datasets_path + 'x_test_offline.txt',dtype=float)
	y_test_offline = np.loadtxt(datasets_path + 'y_test_offline.txt',dtype=float)

	print('Finishing loading data!')

	x_train_cpu = torch.from_numpy(x_train_offline.reshape(-1,1,input_size)).float()
	y_train_cpu = torch.from_numpy(y_train_offline.reshape(-1,1,output_size)).float()
	train_data = CustomDataset(x_train_cpu, y_train_cpu,device)

	x_test_cpu = torch.from_numpy(x_test_offline.reshape(-1,1,input_size)).float()
	y_test_cpu = torch.from_numpy(y_test_offline.reshape(-1,1,output_size)).float()
	test_data = CustomDataset(x_test_cpu, y_test_cpu, device)

	x_test_offline = torch.from_numpy(x_test_offline.reshape(-1,1,input_size)).float().to(device, non_blocking=True)
	y_test_offline = torch.from_numpy(y_test_offline.reshape(-1,1,output_size)).float().to(device, non_blocking=True)

	gpu_device_args = {'num_workers': 4, 'pin_memory': True, 'prefetch_factor': 2} if device=='cuda:1' or device=='cuda:0' else {}
	train_loader = DataLoader(dataset=train_data, batch_size=batch_size, shuffle=True, **gpu_device_args)
	val_loader = DataLoader(dataset=test_data, batch_size=int(batch_size/10), shuffle=False, **gpu_device_args)
	train_step = make_train_step(net1, loss_func, optimizer)

	# Forward pass 
	for t in range(epoch_): 
		for x_batch, y_batch in train_loader:
			net1.train()
			x_batch = x_batch.to(device)
			y_batch = y_batch.to(device)
			loss1 = train_step(x_batch, y_batch)
			loss = loss1
			loss_values.append(loss)
		
		with torch.no_grad():
			for x_val, y_val in val_loader:
				x_val = x_val.to(device)
				y_val = y_val.to(device)
				net1.eval()
				yhat = net1(x_val)
				val_loss = loss_func(y_val, yhat)
				accuracy1 = val_loss
				accuracy = accuracy1
				valoss_values.append(accuracy.item())

		writer.add_scalar('Loss/train',loss,t)
		writer.add_scalar('Loss/test',accuracy,t)

		if t % step_size_ == 0:
			scheduler.step()

		if t % 100 == 0:
			print('Epoch {}/{}'.format(t, epoch_ - 1))
			print('-' * 10)
			print('loss', loss)
			print('acc', accuracy)

		if t % 1000 == 0:
			save(net1, save_path, t)
			torch.onnx.export(net1, x_test_offline, save_path + "mlp_wrap_" + str(t) + \
				"_loss_" + str(accuracy.item()) + ".onnx", \
				export_params=True, opset_version=9, do_constant_folding=True, input_names=['input'], output_names=['output'])   
		
		if accuracy.item() < target_loss:
			save(net1, satisfied_results_path, t)
			torch.onnx.export(net1, x_test_offline, satisfied_results_path + "mlp_wrap_" + str(t) + \
				"_loss_" + str(accuracy.item()) + ".onnx", \
				export_params=True, opset_version=9, do_constant_folding=True, input_names=['input'], output_names=['output'])   

	writer.close()

if __name__ =='__main__':
	parser = argparse.ArgumentParser()
	parser.add_argument('--input', '-i', nargs="?", const=8, default=8, type=int)
	parser.add_argument('--layers', '-l', nargs="?", const=6, default=6, type=int)
	parser.add_argument('--nodes', '-n', nargs="?", const=256, default=256, type=int)
	parser.add_argument('--learning_rate', '-lr', nargs="?", const=0.0001, default=0.0001, type=float)
	parser.add_argument('--epochs', '-e', nargs="?", const=10000, default=10000, type=int)
	parser.add_argument('--train_size', '-train', nargs="?", const=30000, default=30000, type=int)
	parser.add_argument('--test_size', '-test', nargs="?", const=3000, default=3000, type=int)
	parser.add_argument('--batch_size', '-b', nargs="?", const=1000, default=1000, type=int)
	parser.add_argument('--step_size', '-s', nargs="?", const=1000, default=1000, type=int)
	parser.add_argument('--gamma', '-g', nargs="?", const=0.1, default=0.1, type=float)
	parser.add_argument('--use_gpu_device', '-d', nargs="?", const=0, default=-1, type=int)
	parser.add_argument('--threshold', '-th', nargs="?", const=0.01, default=0.01, type=float)
	parser.add_argument('--target_loss', '-tl', nargs="?", const=8e-7, default=8e-7, type=float)

	args = parser.parse_args()
	if args.use_gpu_device >= 0 and torch.cuda.is_available():
		device = torch.device('cuda:{}'.format(args.use_gpu_device))
	else:
		device = torch.device('cpu')

	input_size = args.input
	n_layers = args.layers
	learning_rate = args.learning_rate
	hidden_size = args.nodes
	n_train_samples = args.train_size
	n_test_samples = args.test_size
	epoch_ = args.epochs
	gamma_ = args.gamma
	step_size_ = args.step_size
	threshold = args.threshold
	batch_size = args.batch_size
	target_loss = args.target_loss
	output_size = 3

	# Specify the path where you save your datasets
	datasets_path = os.path.dirname(os.path.abspath(__file__)) + "\\"
	# Specify the path where you want to save the results
	save_path = os.path.dirname(os.path.abspath(__file__)) + "\\"
	satisfied_results_path = save_path + "satisfied_results\\"
	# Specify the path where you want to save your logs
	log_path = "logs\\"

	try:
		os.mkdir(save_path)
		os.mkdir(datasets_path)
		os.mkdir(satisfied_results_path)
	except OSError as error:
		print(error)

	writer = SummaryWriter(log_dir=log_path)

	net1 = Feedforward(input_size, hidden_size, output_size, n_layers, nn.Tanh)
	net1 = net1.to(device)
	net1.train()
	train(net1, save_path)