function driver(sceneID,integrator,batch)
% driver Integrates the scene forward in time
%
% Integrators:
%  1: EULER
%  2: BDF1
%  3: SDIRK2
% 
%{
% To run in batch mode, run the following:
clear; clc;
for sceneID = 0 : 6
	for integrator = 1 : 3
		driver(sceneID,integrator,true)
	end
end
% To check the simulations visually, run the following:
clear; clc;
for sceneID = 0 : 6
	driver(sceneID,1,false)
end
%}

if nargin < 1
	sceneID = 1;
end
if nargin < 2
	integrator = 1;
end
if nargin < 3
	batch = false;
end
switch integrator
	case 1
		iname = 'EULER';
	case 2
		iname = 'BDF1';
	case 3
		iname = 'SDIRK2';
	otherwise
		iname = 'UNKNOWN';
end

% Create scene
scene = scenes(sceneID,integrator);
scene.init();

% Exit if this integrator is not supported
if isnan(scene.Vexpected(integrator))
	fprintf('Integrator not supported!\n')
	return;
end

% Print info
fprintf('(%d) ''%s'' [%s]: tEnd=%.1f, nsteps=%d, nr=%d, nm=%d, na=%d\n',...
	sceneID,scene.name,iname,scene.tEnd,scene.nsteps,...
	redmax.Scene.countR(),redmax.Scene.countM(),redmax.Scene.countA());

% Draw and initialize video
if batch
	scene.drawHz = 0;
	scene.plotH = false;
end
scene.draw();

% Simulation loop
simLoop(scene);

% Plot energies
scene.plotEnergies();

end

%%
function simLoop(scene)
jroot = scene.joints{1};
h = scene.h;
nsteps = scene.nsteps;

% Integrate
for k = 0 : nsteps-1
	% Save old state
	[q0,qdot0] = jroot.getQ();
	jroot.setQ0(q0,qdot0);

	% Compute new state
	switch scene.integrator
		case 1 % EULER
			[q1,qdot1] = euler(scene);
		case 2 % BDF1
			q1 = q0 + h*qdot0; % initial guess
			q1 = solver(@(q1)evalBDF1(q1,q0,qdot0,scene),q1);
			qdot1 = (q1 - q0)/h;
		case 3 % SDIRK2
			a = (2 - sqrt(2))/2;
			qa = q0 + a*h*qdot0; % initial guess
			[qa,H] = solver(@(qa)evalSDIRK2a(qa,q0,qdot0,scene),qa);
			qdota = (qa - q0)/(a*h);
			q1 = qa + (1-a)*h*qdota; % initial guess
			q1 = solver(@(q1)evalSDIRK2b(q1,q0,qdot0,qa,qdota,scene),q1,H);
			qdot1 = (q1 - q0 - (1-a)*h*qdota)/(a*h);
		otherwise
			error('Unknown integrator');
	end

	% Save new state
	Jmr = jroot.computeJacobian();
	Jar = computeMuscleJacobian(scene,Jmr);
	qadot1 = Jar*qdot1;
	scene.update(q1,qdot1,qadot1);
	
	% Update time and step
	scene.t = scene.t + h;
	scene.k = k + 1;
	
	% End of step
	scene.draw();
	scene.saveHistory();
end

end

%%
function [q1,qdot1] = euler(scene)
broot = scene.bodies{1};
jroot = scene.joints{1};
mroot = scene.muscles{1};

h = scene.h;
[q0,qdot0] = jroot.getQ();

% Mm: maximal inertia matrix
% fm: maximal force vector (eg gravity)
% Km: maximal stiffness matrix (eg springs, unused in matlab-simple)
% Dm: maximal damping matrix (eg body damping)
% fr: reduced force vector (eg external joint torques)
% Kr: reduced stiffness matrix (eg joint stiffness)
% Dr: reduced damping matrix (eg joint damping)
% Jmr, dJmr: redmax Jacobian and its time derivative
[Mm,fm] = broot.computeMassGrav(scene.grav);
[fm,Dm] = broot.computeForceDamping(fm);
[fr,Kr] = jroot.computeForceStiffness();
[fr,Dr] = jroot.computeForceDamping(fr);
[Jmr,dJmr] = jroot.computeJacobian();

% Muscle mass, force, and Jacobian
[Ma,fa] = mroot.computeMassGrav(scene.grav);
fm = mroot.computeForceHMM(fm);
[Jar,dJar] = computeMuscleJacobian(scene,Jmr,dJmr);

% Bones
Mr_beta = Jmr'*Mm*Jmr;
frtilde_beta = Mr_beta*qdot0 + h*Jmr'*(fm - Mm*dJmr*qdot0);

% Muscles
Mr_alpha = Jar'*Ma*Jar;
frtilde_alpha = Mr_alpha*qdot0 + h*Jar'*(fa - Ma*dJar*qdot0);

% Add up all, including implicit forces
Mr_all = Mr_beta + Mr_alpha;
LHS = Mr_all + Jmr'*h*Dm*Jmr - h*Dr - h*h*Kr;
rhs = h*fr + frtilde_beta + frtilde_alpha;

% Constraints
[lb,ub] = jroot.computeConstraints();

% Take a time step
if sum(isinf(lb)) + sum(isinf(ub)) == 2*length(qdot0)
	% No inequality constraints
	qdot1 = LHS\rhs;
else
	% With inequality constraints
	baum = 0.1/h; % Baumgarte
	opts = optimoptions(@quadprog,'Display','off');
	qdot1 = quadprog(0.5*(LHS+LHS'),-rhs,[],[],[],[],baum*lb,baum*ub,[],opts);
end
q1 = q0 + h*qdot1;

% Debug: for emulating noisy NN
mroot.setDebugRngSeed(scene.k);

end

%%
function [Jar,dJar] = computeMuscleJacobian(scene,Jmr,dJmr)
% Jar, dJar: muscle Jacobian and its time derivative
nr = redmax.Scene.countR();
na = redmax.Scene.countA();
jroot = scene.joints{1};
mroot = scene.muscles{1};
[q0,qdot0] = jroot.getQ();

useFD = false; % Use finite differencing for everything (useful for debugging)
e = 1e-4; % Can't be too small because of the wrapping optimization

if na == 0
	Jar = zeros(0,nr);
	dJar = zeros(0,nr);
	return;
end

if nargout == 1
	% Only Jar
	if useFD
		% DEBUGGING: Use finite differencing for everything
		% TODO: This doesn't work with NN muscles
		% Fill Jar with muscle mass positions at the current joint angles
		Jar = mroot.computeJacobianFD(); %#ok<UNRCH> 
		% Re-compute the muscle mass positions at perturbed joint angles
		for idxR = 1 : nr
			q0_ = q0;
			q0_(idxR) = q0_(idxR) + e;
			scene.updateQuick(q0_);
			% Calling this method with arguments will generate a finite
			% differenced Jacobian.
			Jar = mroot.computeJacobianFD(idxR,e,Jar);
		end
		scene.updateQuick(q0);
	else
		% Type I, II, and III muscles
		Jam = mroot.computeJacobian();
		Jar = Jam*Jmr;
		% Type III.r muscles (deprecated)
		[JamNN,JarNN] = mroot.computeJacobianNN();
		if norm(JamNN,'fro') > 0
			Jar = Jar + JamNN*Jmr + JarNN;
		end
	end
else
	% Jar and dJar
	if useFD
		% DEBUGGING: Use finite differencing for everything
		% TODO: This doesn't work with NN muscles
		% Fill Jar with muscle mass positions at the current joint angles
		Jar = mroot.computeJacobianFD(); %#ok<UNRCH> 
		% Re-compute the muscle mass positions at perturbed joint angles
		for idxR = 1 : nr
			q0_ = q0;
			q0_(idxR) = q0_(idxR) + e;
			scene.updateQuick(q0_);
			% Calling this method with arguments will generate a finite
			% differenced Jacobian.
			Jar = mroot.computeJacobianFD(idxR,e,Jar);
		end
		
		% Repeat for Jdot
		q1 = q0 + e*qdot0;
		scene.updateQuick(q1);
		Jar_ = mroot.computeJacobianFD();
		for idxR = 1 : nr
			q1_ = q1;
			q1_(idxR) = q1_(idxR) + e;
			scene.updateQuick(q1_);
			Jar_ = mroot.computeJacobianFD(idxR,e,Jar_);
		end
		dJar = (Jar_ - Jar)/e;
		scene.updateQuick(q0);
	else
		% Type I, II, and III muscles
		[Jam,dJam] = mroot.computeJacobian();
		Jar = Jam*Jmr;
		dJar = dJam*Jmr + Jam*dJmr;

		% Type III.r muscles (deprecated)
		[JamNN,JarNN] = mroot.computeJacobianNN();
		if norm(JamNN,'fro') > 0
			Jar = Jar + JamNN*Jmr + JarNN;
			q0_ = q0 + e*qdot0;
			mroot.storeNN(); % save muscle NN
			scene.updateQuick(q0_); % update joints and muscles
			[JamNN_,JarNN_] = mroot.computeJacobianNN();
			jroot.setQ(q0); % set joint dofs
			jroot.updateQuick(); % update joints
			mroot.restoreNN(); % restore muscle NN
			dJamNN = (JamNN_ - JamNN)/e;
			dJarNN = (JarNN_ - JarNN)/e;
			dJar = dJar + dJamNN*Jmr + JamNN*dJmr + dJarNN;
		end
	end
end

end

%%
function g = evalBDF1(q1,q0,qdot0,scene)
broot = scene.bodies{1};
jroot = scene.joints{1};
mroot = scene.muscles{1};

h = scene.h;
qdot1 = (q1 - q0)/h;
scene.update(q1,qdot1);
dq0 = q1 - q0 - h*qdot0;

[Mm,fm] = broot.computeMassGrav(scene.grav);
fm = broot.computeForceDamping(fm);
fr = jroot.computeForceStiffness();
[fr,Dr] = jroot.computeForceDamping(fr); %#ok<ASGLU> 
[Jmr,dJmr] = jroot.computeJacobian();
[Ma,fa] = mroot.computeMassGrav(scene.grav);
[Jar,dJar] = computeMuscleJacobian(scene,Jmr,dJmr);

Mr_beta  = Jmr'*Mm*Jmr;
Mr_alpha = Jar'*Ma*Jar;
fr_beta  = Jmr'*(fm - Mm*dJmr*qdot1);
fr_alpha = Jar'*(fa - Ma*dJar*qdot1);

M = Mr_beta + Mr_alpha;
f = fr_beta + fr_alpha + fr;
g = M*dq0 - (h*h)*f;

end

%%
function g = evalSDIRK2a(qa,q0,qdot0,scene)
broot = scene.bodies{1};
jroot = scene.joints{1};
mroot = scene.muscles{1};

a = (2 - sqrt(2))/2;
h = scene.h;
ah = a*h;
qdota = (qa - q0)/ah;
scene.update(qa,qdota);
dqa = qa - q0 - ah*qdot0;

[Mm,fm] = broot.computeMassGrav(scene.grav);
fm = broot.computeForceDamping(fm);
fr = jroot.computeForceStiffness();
[fr,Dr] = jroot.computeForceDamping(fr); %#ok<ASGLU> 
[Jmr,dJmr] = jroot.computeJacobian();
[Ma,fa] = mroot.computeMassGrav(scene.grav);
[Jar,dJar] = computeMuscleJacobian(scene,Jmr,dJmr);

Mr_beta  = Jmr'*Mm*Jmr;
Mr_alpha = Jar'*Ma*Jar;
fr_beta  = Jmr'*(fm - Mm*dJmr*qdota);
fr_alpha = Jar'*(fa - Ma*dJar*qdota);

M = Mr_beta + Mr_alpha;
f = fr_beta + fr_alpha + fr;

g = M*dqa - (ah*ah)*f;
end

%%
function g = evalSDIRK2b(q1,q0,qdot0,qa,qdota,scene) %#ok<INUSL> 
broot = scene.bodies{1};
jroot = scene.joints{1};
mroot = scene.muscles{1};

a = (2 - sqrt(2))/2;
h = scene.h;
ah = a*h;
qdot1 = (q1 - q0 - (1-a)*h*qdota)/ah;
scene.update(q1,qdot1);
dq1 = q1 - q0 - (2*a-1)*h*qdot0 - (2-2*a)*h*qdota;

[Mm,fm] = broot.computeMassGrav(scene.grav);
fm = broot.computeForceDamping(fm);
fr = jroot.computeForceStiffness();
[fr,Dr] = jroot.computeForceDamping(fr); %#ok<ASGLU> 
[Jmr,dJmr] = jroot.computeJacobian();
[Ma,fa] = mroot.computeMassGrav(scene.grav);
[Jar,dJar] = computeMuscleJacobian(scene,Jmr,dJmr);

Mr_beta  = Jmr'*Mm*Jmr;
Mr_alpha = Jar'*Ma*Jar;
fr_beta  = Jmr'*(fm - Mm*dJmr*qdota);
fr_alpha = Jar'*(fa - Ma*dJar*qdota);

M = Mr_beta + Mr_alpha;
f = fr_beta + fr_alpha + fr;

g = M*dq1 - (ah*ah)*f;
end

%%
function [g,H] = getLinearSystem(evalFcn,x)
n = length(x);

g = evalFcn(x);
e = 1e-4;
H = zeros(n,n);
for i = 1 : length(x)
	x_ = x;
	x_(i) = x_(i) + e;
	g_ = evalFcn(x_);
	H(:,i) = (g_ - g)/e;
end
end

%%
function [x,H] = solver(evalFcn,xInit,H)
% Root finding: solve for g(x) = 0.

if nargin < 3
	H = [];
end

tol = 1e-5;
iterMax = 10*length(xInit);
dxMax = 1e3;

x = xInit;
n = length(x);
m = 5; % L-BFGS

iter = 1;
while true
	if iter == 1
		% Initial Hessian
		if isempty(H)
			[g,H] = getLinearSystem(evalFcn,x);
		else
			g = evalFcn(x);
		end
		% Search direction
		dx = -(H\g);
		% Initialize L-BFGS
		lbfgs = lbfgsInit(m,n);
	else
		% L-BFGS update and solve
		g = evalFcn(x);
		lbfgs = lbfgsUpdate(lbfgs,x,x0,g,g0,iter);
		dx = -lbfgsSolve(lbfgs,g,H);
	end
	if norm(dx) > dxMax
		fprintf('Newton diverged\n');
		break;
	end
	x0 = x;
	g0 = g;
	x = x + dx;
	% Convergence check
	if norm(g) < tol
		% Converged
		break;
	end
	if iter >= iterMax
		fprintf('Newton did not converge after %d iterations\n',iterMax);
		break;
	end
	iter = iter + 1;
end
%fprintf('%d\n',iter);
end

%%
function lbfgs = lbfgsInit(m,n)
lbfgs.m = m;
lbfgs.s = zeros(n,m);
lbfgs.y = zeros(n,m);
lbfgs.rho = zeros(1,m);
lbfgs.ii = zeros(1,0);
end

%%
function lbfgs = lbfgsUpdate(lbfgs,x,x0,g,g0,iter)
% Store the L-BFGS vectors into an array and index them in a
% circular fashion. For example, with m=3, we get:
% iter    1     2     3     4     5     6     7
% imax    -     2     3     1     2     3     1
% ii      []   [2]  [23]  [231] [312] [123] [231]
imax = mod(iter-1,lbfgs.m)+1;
if iter - 1 <= lbfgs.m
	lbfgs.ii(end+1) = imax; %#ok<*AGROW>
else
	lbfgs.ii = [lbfgs.ii(2:end), imax]; % shift
end
si = x - x0;
yi = g - g0;
lbfgs.s(:,imax) = si;
lbfgs.y(:,imax) = yi;
lbfgs.rho(imax) = 1/(si'*yi);
end

%%
function X = lbfgsSolve(lbfgs,B,H)
% L-BFGS on all the columns of B
nrows = lbfgs.m;
ncols = size(B,2);
a = zeros(nrows,ncols);
for k = length(lbfgs.ii) : -1 : 1
	i = lbfgs.ii(k);
	si = lbfgs.s(:,i);
	yi = lbfgs.y(:,i);
	a(i,:) = lbfgs.rho(i)*(si'*B);
	B = B - yi*a(i,:);
end
if nargin < 3
	Hinv = speye(size(B,1));
	if ~isempty(lbfgs.ii)
		si = lbfgs.s(:,lbfgs.ii(end));
		yi = lbfgs.y(:,lbfgs.ii(end));
		Hinv = ((si'*yi)/(yi'*yi))*Hinv;
	end
	X = Hinv*B;
else
	X = H\B;
end
for k = 1 : length(lbfgs.ii)
	i = lbfgs.ii(k);
	yi = lbfgs.y(:,i);
	si = lbfgs.s(:,i);
	ai = a(i,:);
	bi = lbfgs.rho(i)*(yi'*X);
	X = X + si*(ai - bi);
end
end
